import datetime
import os
import logging
from collections import defaultdict
from copy import deepcopy
from typing import Dict, Tuple, TYPE_CHECKING

import pytz

from aim.sdk.configs import AIM_ENABLE_TRACKING_THREAD
from aim.sdk.num_utils import is_number, convert_to_py_number
from aim.sdk.utils import get_object_typename, check_types_compatibility

from aim.storage.hashing import hash_auto
from aim.storage.context import Context
from aim.storage.object import CustomObject
from aim.storage.types import AimObject

if TYPE_CHECKING:
    from aim.sdk import Run

logger = logging.getLogger(__name__)

STEP_HASH_FUNCTIONS = {
    1: lambda s: s,
    2: lambda s: hash_auto(s)
}


class SequenceInfo:
    def __init__(self):
        self.initialized = False
        self.count = None
        self.dtype = None
        self.version = None
        self.val_view = None
        self.step_view = None
        self.epoch_view = None
        self.time_view = None
        self.record_max_length = None
        self.step_hash_fn = None


Selector = Tuple[int, str]


class RunTracker:
    _idx_to_ctx: Dict[int, Context] = dict()
    _track_warning_shown = False

    @classmethod
    def track_rate_warn(cls):
        if not cls._track_warning_shown:
            # TODO [AT] add link to FAQ section in docs.
            logger.warning('Tracking task queue is almost full which might cause performance degradation. '
                           'Consider tracking at lower pace.')
            cls._track_warning_shown = True

    def __init__(self, run: 'Run'):
        self.meta_tree = run.meta_tree
        self.read_only = run.read_only
        self.contexts: Dict[Context, int] = dict()

        if not self.read_only:
            # remote tracking creates dedicated thread for tracking, so don't need to create another one here
            self._non_blocking = os.getenv(AIM_ENABLE_TRACKING_THREAD, False) and not run.repo.is_remote_repo

            self.hash = run.hash
            self.repo = run.repo
            self.meta_run_tree = run.meta_run_tree
            self.series_run_trees = run.series_run_trees
            self.sequence_infos: Dict[Selector, SequenceInfo] = defaultdict(SequenceInfo)

            self._preload_sequence_infos()

    def idx_to_ctx(self, idx):
        ctx = RunTracker._idx_to_ctx.get(idx)
        if ctx is not None:
            return ctx
        ctx = Context(self.meta_tree['contexts', idx])
        RunTracker._idx_to_ctx[idx] = ctx
        self.contexts[ctx] = idx
        return ctx

    def __call__(
        self,
        value,
        name: str = None,
        step: int = None,
        epoch: int = None,
        *,
        context: AimObject = None,
    ):
        assert not self.read_only
        # since worker might be lagging behind, we want to log the timestamp of run.track() call,
        # not the actual implementation execution time.
        track_time = datetime.datetime.now(pytz.utc).timestamp()

        if self._non_blocking:
            val = deepcopy(value)
            self.repo.tracking_queue.register_task(
                self._track, val, track_time, name, step, epoch, context=context) or self.track_rate_warn()
        else:
            self._track(value, track_time, name, step, epoch, context=context)

    def _track(
        self,
        value,
        track_time: float,
        name: str = None,
        step: int = None,
        epoch: int = None,
        *,
        context: AimObject = None,
    ):
        if context is None:
            context = {}

        values = self._normalized_values(value, name)
        with self.repo.atomic_track(self.hash):
            ctx = Context(context)
            for name, val in values.items():
                seq_info = self.sequence_infos[ctx.idx, name]
                if not seq_info.initialized:
                    self._init_sequence_info(ctx.idx, name, val)
                step = step if step is not None else seq_info.count

                self._update_context_data(ctx)
                self._update_sequence_info(ctx.idx, name, val, step)
                self._add_value(seq_info, val, step, epoch, track_time)

    def _preload_sequence_infos(self):
        for ctx_id, traces in self.meta_run_tree.get('traces', {}).items():
            for name in traces:
                try:
                    self._load_sequence_info(ctx_id, name)
                except KeyError:
                    pass

    def _load_sequence_info(self, ctx_id: int, name: str):
        # Creates SequenceInfo for the given context and name.
        # It is assumed that there should be at least one tracked value, otherwise the context/name pair
        #  should not appear on Run traces list. Hence it is OK to check the value type and separate metric
        #  sequences from other object sequences.

        seq_info = self.sequence_infos[(ctx_id, name)]
        assert not seq_info.initialized

        try:
            seq_info.version = self.meta_run_tree['traces', ctx_id, name, 'version']
        except KeyError:
            self.meta_run_tree['traces', ctx_id, name, 'version'] = seq_info.dtype = 1
        try:
            seq_info.dtype = self.meta_run_tree['traces', ctx_id, name, 'dtype']
        except KeyError:
            self.meta_run_tree['traces', ctx_id, name, 'dtype'] = seq_info.dtype = 'float'
        seq_info.step_hash_fn = STEP_HASH_FUNCTIONS[seq_info.version]

        series_tree = self.series_run_trees[seq_info.version]
        if seq_info.version == 2:
            seq_info.step_view = series_tree.subtree((ctx_id, name)).array('step', dtype='int64')
        seq_info.val_view = series_tree.subtree((ctx_id, name)).array('val')
        seq_info.epoch_view = series_tree.subtree((ctx_id, name)).array('epoch', dtype='int64')
        seq_info.time_view = series_tree.subtree((ctx_id, name)).array('time', dtype='int64')
        seq_info.count = self.meta_run_tree['traces', ctx_id, name, 'last_step'] + 1
        seq_info.record_max_length = self.meta_run_tree.get(('traces', ctx_id, name, 'record_max_length'), 0)

        seq_info.initialized = True

    def _init_sequence_info(self, ctx_id: int, name: str, val):
        # this method is used in the `run.track()`, so please use only write-only instructions
        seq_info = self.sequence_infos[ctx_id, name]
        assert not seq_info.initialized

        # new SequenceInfo, initialize
        # the subtree().array().allocate() method is write-only
        seq_info.count = 0
        seq_info.record_max_length = 0
        seq_info.dtype = None
        seq_info.version = 2 if get_object_typename(val) in ('int', 'float') else 1
        seq_info.step_hash_fn = STEP_HASH_FUNCTIONS[seq_info.version]

        series_tree = self.series_run_trees[seq_info.version]
        if seq_info.version == 2:
            seq_info.step_view = series_tree.subtree((ctx_id, name)).array('step', dtype='int64').allocate()
        seq_info.val_view = series_tree.subtree((ctx_id, name)).array('val').allocate()
        seq_info.epoch_view = series_tree.subtree((ctx_id, name)).array('epoch', dtype='int64').allocate()
        seq_info.time_view = series_tree.subtree((ctx_id, name)).array('time', dtype='int64').allocate()
        seq_info.initialized = True
        return seq_info

    def _update_sequence_info(self, ctx_id: int, name: str, val, step: int):
        # this method is used in the `run.track()`, so please use only write-only instructions
        seq_info = self.sequence_infos[ctx_id, name]
        assert seq_info.initialized

        dtype = get_object_typename(val)

        if seq_info.dtype is not None:
            def update_trace_dtype(old_dtype: str, new_dtype: str):
                logger.warning(f'Updating sequence \'{name}\' data type from {old_dtype} to {new_dtype}.')
                self.meta_tree['traces_types', new_dtype, ctx_id, name] = 1
                self.meta_run_tree['traces', ctx_id, name, 'dtype'] = new_dtype
                seq_info.dtype = new_dtype

            compatible = check_types_compatibility(dtype, seq_info.dtype, update_trace_dtype)
            if not compatible:
                raise ValueError(f'Cannot log value \'{val}\' on sequence \'{name}\'. Incompatible data types.')

        if seq_info.count == 0:
            self.meta_tree['traces_types', dtype, ctx_id, name] = 1
            self.meta_run_tree['traces', ctx_id, name, 'dtype'] = dtype
            self.meta_run_tree['traces', ctx_id, name, 'version'] = seq_info.version
            self.meta_run_tree['traces', ctx_id, name, 'first_step'] = step
            seq_info.dtype = dtype

        if step >= seq_info.count:
            self.meta_run_tree['traces', ctx_id, name, 'last'] = val
            self.meta_run_tree['traces', ctx_id, name, 'last_step'] = step
            seq_info.count = step + 1

        if isinstance(val, (tuple, list)):
            record_max_length = max(seq_info.record_max_length, len(val))
            self.meta_run_tree['traces', ctx_id, name, 'record_max_length'] = record_max_length
            seq_info.record_max_length = record_max_length

    def _update_context_data(self, ctx: Context):
        if ctx not in self.contexts:
            self.meta_tree['contexts', ctx.idx] = ctx.to_dict()
            self.meta_run_tree['contexts', ctx.idx] = ctx.to_dict()
            self.contexts[ctx] = ctx.idx
            self._idx_to_ctx[ctx.idx] = ctx

    def _add_value(self, seq_info, val, step, epoch, track_time):
        step_hash = seq_info.step_hash_fn(step)
        if seq_info.step_view is not None:
            seq_info.step_view[step_hash] = step
        seq_info.val_view[step_hash] = val
        seq_info.epoch_view[step_hash] = epoch
        seq_info.time_view[step_hash] = track_time

    @staticmethod
    def _normalized_values(value, name):
        def _normalize_single_value(val):
            if is_number(val):
                return convert_to_py_number(val)
            elif isinstance(val, (CustomObject, list, tuple)):
                return val
            else:
                raise ValueError(f'Input type {type(value)} is neither python number nor AimObject')

        if isinstance(value, dict):
            if name is not None:
                raise ValueError('\'name\' should be None when tracking values dictionary.')
            return {str(k): _normalize_single_value(v) for k, v in value.items()}
        else:
            if name is None:
                raise ValueError('\'name\' should not be None.')
            return {name: _normalize_single_value(value)}



from pydantic import BaseModel
from typing import List, Optional, Dict
from uuid import UUID


class ExperimentCreateIn(BaseModel):
    name: str


class ExperimentUpdateIn(BaseModel):
    name: Optional[str] = ''
    description: Optional[str] = ''
    archived: Optional[bool] = None


class ExperimentGetOut(BaseModel):
    id: UUID
    name: str
    description: Optional[str] = ''
    run_count: int
    archived: bool
    creation_time: Optional[float]


ExperimentListOut = List[ExperimentGetOut]


class ExperimentUpdateOut(BaseModel):
    id: UUID
    status: str = 'OK'


class ExperimentGetRunsOut(BaseModel):
    class Run(BaseModel):
        run_id: str
        name: str
        creation_time: float
        end_time: Optional[float]
        archived: bool

    id: UUID
    runs: List[Run]


class ExperimentActivityApiOut(BaseModel):
    num_runs: int
    num_archived_runs: int
    num_active_runs: int
    activity_map: Dict[str, int] = {"2021-01-01": 54}



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: base.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='base.proto',
  package='artifacts.proto',
  syntax='proto3',
  serialized_pb=_b('\n\nbase.proto\x12\x0f\x61rtifacts.proto\"a\n\nBaseRecord\x12\x0c\n\x04step\x18\x01 \x01(\x05\x12\x11\n\ttimestamp\x18\x02 \x01(\x03\x12\x10\n\x08\x61rtifact\x18\x03 \x01(\x0c\x12\r\n\x05\x65poch\x18\x04 \x01(\x05\x12\x11\n\thas_epoch\x18\x05 \x01(\x08\x62\x06proto3')
)




_BASERECORD = _descriptor.Descriptor(
  name='BaseRecord',
  full_name='artifacts.proto.BaseRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='step', full_name='artifacts.proto.BaseRecord.step', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='artifacts.proto.BaseRecord.timestamp', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='artifact', full_name='artifacts.proto.BaseRecord.artifact', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='epoch', full_name='artifacts.proto.BaseRecord.epoch', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='has_epoch', full_name='artifacts.proto.BaseRecord.has_epoch', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=31,
  serialized_end=128,
)

DESCRIPTOR.message_types_by_name['BaseRecord'] = _BASERECORD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BaseRecord = _reflection.GeneratedProtocolMessageType('BaseRecord', (_message.Message,), dict(
  DESCRIPTOR = _BASERECORD,
  __module__ = 'base_pb2'
  # @@protoc_insertion_point(class_scope:artifacts.proto.BaseRecord)
  ))
_sym_db.RegisterMessage(BaseRecord)


# @@protoc_insertion_point(module_scope)



import os

from fastapi import FastAPI
from fastapi.exceptions import HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import JSONResponse

from aim.sdk.configs import get_aim_repo_name
from aim.web.configs import AIM_PROFILER_KEY
from aim.web.middlewares.profiler import PyInstrumentProfilerMiddleware
from aim.web.utils import get_root_path


async def http_exception_handler(request, exc):
    message = str(exc.detail)
    detail = None

    if isinstance(exc.detail, dict):
        message = exc.detail.pop('message', message)
        detail = exc.detail.pop('detail', None)

    response = {'message': message}
    if detail:
        response.update({'detail': detail})
    else:
        response.update({'detail': str(exc)})
    return JSONResponse(response, status_code=exc.status_code)


async def fallback_exception_handler(request, exc):
    response = {
        'message': f'\'{type(exc)}\' exception raised!',
        'detail': str(exc)
    }
    return JSONResponse(response, status_code=500)


def create_app():
    app = FastAPI(title=__name__)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_methods=['GET', 'POST', 'PUT', 'DELETE', 'OPTIONS', 'HEAD'],
        allow_headers=['Origin', 'X-Requested-With',
                       'Content-Type', 'Accept', 'Authorization', 'X-Timezone-Offset'],
        allow_credentials=True,
        max_age=86400
    )

    from aim.web.api.runs.views import runs_router
    from aim.web.api.runs.views import add_api_routes
    from aim.web.api.tags.views import tags_router
    from aim.web.api.experiments.views import experiment_router
    from aim.web.api.dashboard_apps.views import dashboard_apps_router
    from aim.web.api.dashboards.views import dashboards_router
    from aim.web.api.projects.views import projects_router
    from aim.web.api.views import statics_router
    from aim.web.configs import AIM_UI_BASE_PATH

    from aim.web.api.projects.project import Project
    from aim.sdk.index_manager import RepoIndexManager

    # The indexing thread has to run in the same process as the uvicorn app itself.
    # This allows sharing state of indexing using memory instead of process synchronization methods.
    index_mng = RepoIndexManager.get_index_manager(Project().repo)
    index_mng.start_indexing_thread()

    api_app = FastAPI()
    api_app.add_middleware(GZipMiddleware)
    api_app.add_exception_handler(HTTPException, http_exception_handler)
    api_app.add_exception_handler(Exception, fallback_exception_handler)

    if os.environ.get(AIM_PROFILER_KEY) == "1":
        api_app.add_middleware(PyInstrumentProfilerMiddleware,
                               repo_path=os.path.join(get_root_path(), get_aim_repo_name()))

    add_api_routes()

    api_app.include_router(dashboard_apps_router, prefix='/apps')
    api_app.include_router(dashboards_router, prefix='/dashboards')
    api_app.include_router(experiment_router, prefix='/experiments')
    api_app.include_router(projects_router, prefix='/projects')
    api_app.include_router(runs_router, prefix='/runs')
    api_app.include_router(tags_router, prefix='/tags')

    base_path = os.environ.get(AIM_UI_BASE_PATH, '')

    app.mount(f'{base_path}/api', api_app)
    static_files_app = FastAPI()

    static_files_app.include_router(statics_router)
    app.mount(f'{base_path}/', static_files_app)

    return app



import json
import logging
import uuid

from typing import Dict
from pathlib import Path

from aim.ext.notifier.utils import get_empty_config_path


class Config:
    def __init__(self, config_file: Path):
        self._cfg_file: Path = config_file
        self._cfg = {}
        if self.exists():
            self._from_file(self._cfg_file)
        else:
            self._from_file(get_empty_config_path())

    def exists(self) -> bool:
        return self._cfg_file.exists()

    def load(self):
        self._from_file(self._cfg_file)

    def _from_file(self, cfg_file: Path):
        with cfg_file.open() as cfg_fh:
            self._cfg = json.load(cfg_fh)

    def save(self):
        with self._cfg_file.open('w+') as cfg_fh:
            json.dump(self._cfg, cfg_fh, indent=2)

    def dump(self) -> str:
        return json.dumps(self._cfg, indent=2)

    def add(self, config: Dict):
        cfg_id = config['id']
        config['status'] = 'enabled'
        self._cfg['notifications']['notifiers'][cfg_id] = config

    def get(self, cfg_id: uuid.UUID) -> Dict:
        return self.notifiers.get(cfg_id, {})

    def enable(self, cfg_id: uuid.UUID):
        cfg = self.notifiers[cfg_id]
        cfg['status'] = 'enabled'

    def disable(self, cfg_id: uuid.UUID):
        cfg = self.notifiers[cfg_id]
        cfg['status'] = 'disabled'

    def remove(self, cfg_id: uuid.UUID) -> bool:
        if cfg_id in self.notifiers:
            del self.notifiers[cfg_id]
            return True
        return False

    @property
    def notifiers(self) -> Dict:
        return self._cfg['notifications']['notifiers']

    @property
    def log_level(self) -> int:
        return self._cfg['notifications'].get('log_level', logging.WARNING)

    @log_level.setter
    def log_level(self, lvl: int):
        self._cfg['notifications']['log_level'] = lvl



import logging
import os
from pathlib import Path

import aimrocks

from typing import Iterator, Optional, Tuple

from aim.ext.cleanup import AutoClean
from aim.ext.exception_resistant import exception_resistant
from aim.storage.locking import SoftFileLock, NoopLock, DualLock
from aim.storage.types import BLOB
from aim.storage.container import Container, ContainerKey, ContainerValue, ContainerItemsIterator
from aim.storage.prefixview import PrefixView
from aim.storage.containertreeview import ContainerTreeView
from aim.storage.treeview import TreeView


logger = logging.getLogger(__name__)

BLOB_SENTINEL = b''
BLOB_DOMAIN = b'BLOBS\xfe'


class RocksAutoClean(AutoClean):
    PRIORITY = 60

    def __init__(self, instance: 'RocksContainer') -> None:
        """
        Prepare the `RocksContainer` for automatic cleanup.

        Args:
            instance: The `RocksContainer` instance to be cleaned up.
        """
        super().__init__(instance)
        self._lock = None
        self._db = None

    def _close(self):
        """
        Close the RocksDB instances, flush memtables and WAL.
        Finally, release the lock.
        """
        if self._lock is not None:
            if self._db is not None:
                self._db.flush()
                self._db.flush_wal()
                self._db = None
            self._lock.release()
            self._lock = None
        if self._db is not None:
            self._db = None


class RocksContainer(Container):
    """
    TODO Rocks-specific docs
    """
    def __init__(
        self,
        path: str,
        read_only: bool = False,
        wait_if_busy: bool = False,
        **extra_options
    ) -> None:
        self._resources: RocksAutoClean = None

        self.path = Path(path)
        self.read_only = read_only
        self._db_opts = dict(
            create_if_missing=True,
            paranoid_checks=False,
            keep_log_file_num=10,
            skip_stats_update_on_db_open=True,
            skip_checking_sst_file_sizes_on_db_open=True,
            max_open_files=-1,
            write_buffer_size=1024 * 1024,  # 1MB
            db_write_buffer_size=1024 * 1024,  # 1MB
            max_write_buffer_number=1,
            target_file_size_base=64 * 1024 * 1024,  # 64MB
            max_background_compactions=4,
            level0_file_num_compaction_trigger=8,
            level0_slowdown_writes_trigger=17,
            level0_stop_writes_trigger=24,
            num_levels=4,
            max_bytes_for_level_base=512 * 1024 * 1024,  # 512MB
            max_bytes_for_level_multiplier=8,
        )
        self._extra_opts = extra_options
        # opts.allow_concurrent_memtable_write = False
        # opts.memtable_factory = aimrocks.VectorMemtableFactory()
        # opts.table_factory = aimrocks.PlainTableFactory()
        # opts.table_factory = aimrocks.BlockBasedTableFactory(block_cache=aimrocks.LRUCache(67108864))
        # opts.write_buffer_size = 67108864
        # opts.arena_block_size = 67108864

        self._wait_if_busy = wait_if_busy  # TODO implement
        self._lock_path: Optional[Path] = None
        self._progress_path: Optional[Path] = None

        self._resources = RocksAutoClean(self)

        if not self.read_only:
            self.writable_db
        # TODO check if Containers are reopenable

    # The following properties are linked to self._resources to
    # ensure that the resources are closed when the container gone.

    def get_lock_cls(self):
        """Default locking is no-op. Container locking to be handled externally."""
        return NoopLock

    @property
    def _db(self):
        return self._resources._db

    @_db.setter
    def _db(self, value):
        self._resources._db = value

    @property
    def _lock(self):
        return self._resources._lock

    @_lock.setter
    def _lock(self, value):
        self._resources._lock = value

    @property
    def db(self) -> aimrocks.DB:
        if self._db is not None:
            return self._db

        logger.debug(f'opening {self.path} as aimrocks db')
        if not self.read_only:
            lock_path = prepare_lock_path(self.path)
            self._lock_path = lock_path
            timeout = self._extra_opts.get('timeout', 10)
            lock_cls = self.get_lock_cls()
            self._lock = lock_cls(self._lock_path, timeout)
            self._lock.acquire()
        else:
            self.optimize_for_read()

        self._db = aimrocks.DB(str(self.path),
                               aimrocks.Options(**self._db_opts),
                               read_only=self.read_only)

        return self._db

    @property
    def writable_db(self) -> aimrocks.DB:
        db = self.db
        if self._progress_path is None:
            progress_dir = self.path.parent.parent / 'progress'
            progress_dir.mkdir(parents=True, exist_ok=True)
            self._progress_path = progress_dir / self.path.name
            self._progress_path.touch(exist_ok=True)
        return db

    def finalize(self, index: Container):
        """Finalize the Container.

        Store the collection of `(key, value)` records in the :obj:`Container`
        `index` for fast reads.
        """
        if not self._progress_path:
            return

        for k, v in self.items():
            index[k] = v

        self._progress_path.unlink()
        self._progress_path = None

    def close(self):
        """Close all the resources."""
        if self._resources is None:
            return
        self._resources.close()

    def preload(self):
        """Preload the Container in the read mode."""
        self.db

    def tree(self) -> 'TreeView':
        """Return a :obj:`ContainerTreeView` which enables hierarchical view and access
        to the container records.

        This is achieved by prefixing groups and using `PATH_SENTINEL` as a
        separator for keys.

        For example, if the Container contents are:
            `{
                b'e.y': b'012',
                b'meta.x': b'123',
                b'meta.z': b'x',
                b'zzz': b'oOo'
            }`, and the path sentinel is `b'.'` then `tree = container.tree()`
            will behave as a (possibly deep) dict-like object:
            `tree[b'meta'][b'x'] == b'123'`
        """
        return ContainerTreeView(self)

    def get(
        self,
        key: ContainerKey,
        default=None
    ) -> ContainerValue:
        """Returns the value by the given `key` if it exists else `default`.

        The `default` is :obj:`None` by default.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(
        self,
        key: ContainerKey
    ) -> ContainerValue:
        """Returns the value by the given `key`.

        Raises :obj:`KeyError` if the `key` is not found.
        """
        value = self.db.get(key=key)
        if value is None:
            raise KeyError(key)
        if value == BLOB_SENTINEL:
            return self._get_blob(key)
        return value

    def _get_blob_loader(
        self,
        key: ContainerKey
    ):
        def loader() -> bytes:
            data = self[BLOB_DOMAIN + key]
            assert isinstance(data, bytes)
            return data
        return loader

    def _get_blob(
        self,
        key: ContainerKey
    ) -> BLOB:
        return BLOB(loader_fn=self._get_blob_loader(key))

    def _put(
        self,
        key: ContainerKey,
        value: ContainerValue,
        *,
        target: aimrocks.WriteBatch
    ):
        target.put(key, value)

    def _put_blob(
        self,
        key: ContainerKey,
        value: BLOB,
        *,
        target: aimrocks.WriteBatch
    ):
        self._put(key=BLOB_DOMAIN + key,
                  value=bytes(value),
                  target=target)

    def _delete(
        self,
        key: ContainerKey,
        *,
        target: aimrocks.WriteBatch
    ):
        target.delete(key)

    def _delete_blob(
        self,
        key: ContainerKey,
        *,
        target: aimrocks.WriteBatch
    ):
        self._delete(key=BLOB_DOMAIN + key,
                     target=target)

    def _delete_range(
        self,
        begin: ContainerKey,
        end: ContainerKey,
        *,
        target: aimrocks.WriteBatch
    ):
        target.delete_range(begin, end)

    def _delete_blob_range(
        self,
        begin: ContainerKey,
        end: ContainerKey,
        *,
        target: aimrocks.WriteBatch
    ):
        self._delete_range(BLOB_DOMAIN + begin,
                           BLOB_DOMAIN + end,
                           target=target)

    def set(
        self,
        key: ContainerKey,
        value: ContainerValue,
        store_batch: aimrocks.WriteBatch = None
    ):
        """Set a value for given key, optionally store in a batch.

        If `store_batch` is provided, instead of the `(key, value)` being added
        to the collection immediately, the operation is stored in a batch in
        order to be executed in a whole with other write operations.

        See :obj:`RocksContainer.batch` and :obj:`RocksContainer.commit` for
        more details.
        """
        target = self.batch() if store_batch is None else store_batch

        if isinstance(value, BLOB):
            self._put(key=key, value=BLOB_SENTINEL, target=target)
            self._put_blob(key=key, value=value, target=target)
        else:
            self._put(key=key, value=value, target=target)
            self._delete_blob(key=key, target=target)

        if store_batch is None:
            self.commit(target)

    def __setitem__(
        self,
        key: ContainerKey,
        value: ContainerValue
    ):
        """Set a value for given key."""
        self.set(key=key, value=value)

    def delete(
        self,
        key: ContainerKey,
        *,
        store_batch: aimrocks.WriteBatch = None
    ):
        """Delete a key-value record by the given key,
        optionally store in a batch.

        If `store_batch` is provided, instead of the `(key, value)` being added
        to the collection immediately, the operation is stored in a batch in
        order to be executed in a whole with other write operations.

        See :obj:`RocksContainer.batch` and :obj:`RocksContainer.commit` for
        more details.
        """
        target = self.batch() if store_batch is None else store_batch

        self._delete(key, target=target)
        self._delete_blob(key, target=target)

        if store_batch is None:
            self.commit(target)

    def __delitem__(
        self,
        key: ContainerKey
    ) -> None:
        """Delete a key-value record by the given key."""
        self.delete(key)

    def delete_range(
        self,
        begin: ContainerKey,
        end: ContainerKey,
        store_batch: aimrocks.WriteBatch = None
    ):
        """Delete all the records in the given `[begin, end)` key range,
        optionally store in a batch.

        If `store_batch` is provided, instead of the `(key, value)` being added
        to the collection immediately, the operation is stored in a batch in
        order to be executed in a whole with other write operations.

        See :obj:`RocksContainer.batch` and :obj:`RocksContainer.commit` for
        more details.
        """
        target = self.batch() if store_batch is None else store_batch

        self._delete_range(begin, end, target=target)
        self._delete_blob_range(begin, end, target=target)

        if store_batch is None:
            self.commit(target)

    def items(
        self,
        prefix: ContainerKey = b''
    ) -> Iterator[Tuple[ContainerKey, ContainerValue]]:
        """Iterate over all the key-value records in the prefix key range.

        The iteration is always performed in lexiographic order w.r.t keys.
        If `prefix` is provided, iterate only over those records that have key
        starting with the `prefix`.

        For example, if `prefix == b'meta.'`, and the Container consists of:
        `{
            b'e.y': b'012',
            b'meta.x': b'123',
            b'meta.z': b'x',
            b'zzz': b'oOo'
        }`, the method will yield `(b'meta.x', b'123')` and `(b'meta.z', b'x')`

        Args:
            prefix (:obj:`bytes`): the prefix that defines the key range
        """
        return RocksContainerItemsIterator(container=self, prefix=prefix)

    def walk(
        self,
        prefix: ContainerKey = b''
    ):
        """A bi-directional generator to walk over the collection of records on
        any arbitrary order. The `prefix` sent to the generator (lets call it
        a `walker`) seeks for lower-bound key in the collection.

        In other words, if the Container contents are:
        `{
            b'e.y': b'012',
            b'meta.x': b'123',
            b'meta.z': b'x',
            b'zzz': b'oOo'
        }` and `walker = container.walk()` then:
        `walker.send(b'meta') == b'meta.x'`, `walker.send(b'e.y') == b'e.y'`
        """
        it = self.db.iteritems()
        it.seek(prefix)

        while True:
            try:
                key, val = next(it)
            except StopIteration:
                yield None
                break
            jump = yield key
            it.seek(jump)

    def view(
        self,
        prefix: ContainerKey = b''
    ) -> 'Container':
        """Return a view (even mutable ones) that enable access to the container
        but with modifications.

        Args:
            prefix (:obj:`bytes`): the prefix that defines the key range of the
                view-container. The resulting container will share an access to
                only records in the `prefix` key range, but with `prefix`-es
                stripped from them.

                For example, if the Container contents are:
                `{
                    b'e.y': b'012',
                    b'meta.x': b'123',
                    b'meta.z': b'x',
                    b'zzz': b'oOo'
                }`, then `container.view(prefix=b'meta.')` will behave (almost)
                exactly as an Container:
                `{
                    b'x': b'123',
                    b'z': b'x',
                }`
        """
        return PrefixView(prefix=prefix, container=self)

    def batch(
        self
    ) -> aimrocks.WriteBatch:
        """Creates a new batch object to store operations in before executing
        using :obj:`RocksContainer.commit`.

        The operations :obj:`RocksContainer.set`, :obj:`RocksContainer.delete`,
        :obj:`RocksContainer.delete_range` are supported.

        See more at :obj:`RocksContainer.commit`
        """
        return aimrocks.WriteBatch()

    def commit(
        self,
        batch: aimrocks.WriteBatch
    ):
        """Execute the accumulated write operations in the given `batch`.

        The `RocksContainer` features atomic writes for batches.
        """
        self.writable_db.write(batch)

    def next_item(
        self,
        prefix: ContainerKey = b''
    ) -> Tuple[ContainerKey, ContainerValue]:
        """Returns `(key, value)` for the key that comes (lexicographically)
        right after the provided `key`.
        """
        it = self.db.iteritems()
        it.seek(prefix + b'\x00')

        key, value = next(it)

        if not key.startswith(prefix):
            raise KeyError

        if value == BLOB_SENTINEL:
            value = self._get_blob(key)

        return key, value

    def prev_item(
        self,
        prefix: ContainerKey = b''
    ) -> Tuple[ContainerKey, ContainerValue]:
        """Returns `(key, value)` for the key that comes (lexicographically)
        right before the provided `key`.
        """
        it = self.db.iteritems()
        it.seek_for_prev(prefix + b'\xff')

        key, value = it.get()

        if value == BLOB_SENTINEL:
            value = self._get_blob(key)

        return key, value

    def optimize_for_read(self):
        optimize_db_for_read(self.path, self._db_opts, run_compactions=self._extra_opts.get('compaction', False))


@exception_resistant(silent=True)
def optimize_db_for_read(path: Path, options: dict, run_compactions: bool = False):
    """
    This function will try to open rocksdb db in write mode and force WAL files recovery. Once done the underlying
    db will contain .sst files only which will significantly reduce further open and read operations. Further
    optimizations can be done by running compactions but this is a costly operation to be performed online.


    Args:
        path (:obj:`Path`): Path to rocksdb.
        options (:obj:`dict`): options to be passed to aimrocks.DB object __init__.
        run_compactions (:obj:`bool`, optional): Flag used to run rocksdb range compactions. False by default.
    """

    def non_empty_wal():
        for wal_path in path.glob('*.log'):
            if os.path.getsize(wal_path) > 0:
                return True
        return False

    if non_empty_wal():
        lock_path = prepare_lock_path(path)

        with SoftFileLock(lock_path, timeout=0):
            wdb = aimrocks.DB(str(path), aimrocks.Options(**options), read_only=False)
            wdb.flush()
            wdb.flush_wal()
            if run_compactions:
                wdb.compact_range()
            del wdb


def prepare_lock_path(path: Path):
    """
    This function creates the locks directory (if needed) and returns a LOCK file path for given rocksdb `path`.

    Args:
        path (:obj:`Path`): Path to rocksdb.

    Returns:
        path (:obj:`Path`) to lock file for given rocksdb.
    """
    path.parent.mkdir(parents=True, exist_ok=True)
    locks_dir = path.parent.parent / 'locks'
    locks_dir.mkdir(parents=True, exist_ok=True)
    return locks_dir / path.name


class RocksContainerItemsIterator(ContainerItemsIterator):
    def __init__(
        self,
        container: RocksContainer,
        prefix: ContainerKey = b''
    ):
        self.container = container
        self.prefix = prefix
        self.it = self.container.db.iteritems()
        self.it.seek(prefix)

    def next(self):
        item = self.it.next()

        if item is None:
            return None

        key = item[0]
        value = item[1]

        if not key.startswith(self.prefix):
            return None

        if value == BLOB_SENTINEL:
            return key, self.container._get_blob(key)

        return key, value


class LockableRocksContainer(RocksContainer):
    def get_lock_cls(self):
        """Use both Unix file-locks and Soft file-locks to cover all the scenarios and avoid corruptions."""
        return DualLock



import os
import json
import re
import hashlib

from aim.sdk.configs import get_aim_repo_name
from aim.sdk.utils import clean_repo_path
from aim.cli.upgrade._legacy_repo.repo.utils import get_experiment_run_path
from aim.cli.upgrade._legacy_repo.configs import AIM_CONFIG_FILE_NAME, AIM_COMMIT_INDEX_DIR_NAME


class AimRepo:
    # TODO: Refactor repo to have minimal side effects
    WRITING_MODE = 'w'
    READING_MODE = 'r'

    def __init__(self, path=None, repo_branch=None,
                 repo_commit=None,
                 repo_full_path=None,
                 mode=WRITING_MODE):
        self._config = {}
        path = clean_repo_path(path)
        self.path = repo_full_path or os.path.join(path, get_aim_repo_name())
        self.config_path = os.path.join(self.path, AIM_CONFIG_FILE_NAME)
        self.hash = hashlib.md5(self.path.encode('utf-8')).hexdigest()

        self.active_commit = repo_commit or AIM_COMMIT_INDEX_DIR_NAME
        if re.match(r'^[A-Za-z0-9_\-]{2,}$', self.active_commit) is None:
            raise ValueError('run name must be at least 2 characters ' +
                             'and contain only latin letters, numbers, ' +
                             'dash and underscore')

        self.root_path = repo_full_path or path
        self.name = self.root_path.split(os.sep)[-1]

        self.branch_path = None
        self.index_path = None
        self.objects_dir_path = None
        self.media_dir_path = None
        self.records_storage = None
        self.mode = mode

        active_exp = self.config.get('active_branch')

        if repo_branch is not None:
            experiment = repo_branch
        elif active_exp is not None:
            experiment = active_exp
        else:
            experiment = None

        if experiment is not None:
            run_full_path = get_experiment_run_path(self.path,
                                                    experiment,
                                                    self.active_commit)
        else:
            run_full_path = None

        if self.active_commit != AIM_COMMIT_INDEX_DIR_NAME and run_full_path \
                and os.path.exists(run_full_path):
            raise ValueError(('run `{}` already exists' +
                              '').format(self.active_commit))

    def __str__(self):
        return self.path

    @property
    def config(self):
        """
        Config property getter, loads config file if not already loaded and
        returns json object
        """
        if len(self._config) == 0:
            if os.path.isfile(self.config_path):
                with open(self.config_path, 'r') as f:
                    config = json.load(f)
                self._config = config
        return self._config

    def get_records_storage(self, path, mode):
        from aimrecords import Storage
        return Storage(path, mode)

    def list_branches(self) -> list:
        """
        Returns list of existing branches
        """
        if self.config.get('branches') is None:
            return []

        return list(filter(lambda b: b != '',
                           map(lambda b: b.get('name') if b else '',
                               self.config.get('branches'))))

    def list_branch_commits(self, branch):
        """
        Returns list of specified branch commits
        """
        branch_path = os.path.join(self.path, branch.strip())

        commits = []

        if not os.path.isdir(branch_path):
            return commits

        for i in os.listdir(branch_path):
            if os.path.isdir(os.path.join(branch_path, i)) \
                    and i != AIM_COMMIT_INDEX_DIR_NAME:
                commits.append(i)
        return commits



from tests.base import TestBase

from aim.sdk.run import Run
from aim.sdk.sequences.metric import Metric
from aim.storage.context import Context


class TestRunSequenceHomogeneousValues(TestBase):
    def test_different_types_on_different_contexts_and_runs(self):
        run = Run(system_tracking_interval=None)
        # same sequence name, different contexts
        run.track(1., name='numbers', context={'type': 'float'})
        run.track(1, name='numbers', context={'type': 'integer'})

        run2 = Run(system_tracking_interval=None)
        # same sequence name, different runs
        run2.track(1, name='numbers', context={'type': 'float'})

    def test_incompatible_type_during_tracking(self):
        run = Run(system_tracking_interval=None)
        run.track(1., name='numbers', context={})
        with self.assertRaises(ValueError) as cm:
            run.track([1], name='numbers', context={})
        exception = cm.exception
        self.assertEqual('Cannot log value \'[1]\' on sequence \'numbers\'. Incompatible data types.', exception.args[0])

    def test_track_metrics_dict(self):
        run = Run(system_tracking_interval=None)
        empty_context_idx = Context({}).idx
        for index in range(10):
            run.track({'numbers': index, 'squares': index**2})
        metric_infos = run.meta_run_tree.subtree(('traces', empty_context_idx))
        self.assertEqual('int', metric_infos['numbers', 'dtype'])
        self.assertEqual('int', metric_infos['squares', 'dtype'])
        self.assertEqual(9, metric_infos['numbers', 'last'])
        self.assertEqual(81, metric_infos['squares', 'last'])

    def test_track_dicts_name_requirements(self):
        run = Run(system_tracking_interval=None)
        with self.assertRaises(ValueError) as cm:
            run.track({'number': 1}, name='number')
        exception = cm.exception
        self.assertEqual('\'name\' should be None when tracking values dictionary.', exception.args[0])

        with self.assertRaises(ValueError) as cm:
            run.track(1)
        exception = cm.exception
        self.assertEqual('\'name\' should not be None.', exception.args[0])

    def test_incompatible_type_after_tracking_restart(self):
        run = Run(system_tracking_interval=None)
        run_hash = run.hash
        run.track(1., name='numbers', context={})
        run.finalize()
        del run

        new_run = Run(run_hash=run_hash, system_tracking_interval=None)
        with self.assertRaises(ValueError) as cm:
            new_run.track([1], name='numbers', context={})
        exception = cm.exception
        self.assertEqual('Cannot log value \'[1]\' on sequence \'numbers\'. Incompatible data types.', exception.args[0])

    def test_type_compatibility_for_empty_list(self):
        run = Run(system_tracking_interval=None)
        context = {}
        ctx = Context(context)
        seq_name = 'obj_list'

        sequence_info = run.meta_run_tree.subtree(('traces', ctx.idx, seq_name))
        typed_sequences_info = run.meta_tree.subtree('traces_types')

        run.track([], name=seq_name, context=context)
        self.assertEqual('list', sequence_info['dtype'])
        self.assertEqual(1, typed_sequences_info['list', ctx.idx, seq_name])
        self.assertIsNone(typed_sequences_info.get(('list(float)', ctx.idx, seq_name), None))

        run.track([], name=seq_name, context=context)
        self.assertEqual('list', sequence_info['dtype'])
        self.assertIsNone(typed_sequences_info.get(('list(float)', ctx.idx, seq_name), None))

        run.track([1.], name=seq_name, context=context)
        self.assertEqual('list(float)', sequence_info['dtype'])
        self.assertEqual(1, typed_sequences_info['list(float)', ctx.idx, seq_name])

        run.track([], name=seq_name, context=context)
        self.assertEqual('list(float)', sequence_info['dtype'])

        with self.assertRaises(ValueError) as cm:
            run.track([5], name=seq_name, context=context)
        exception = cm.exception
        self.assertEqual(
            f'Cannot log value \'{[5]}\' on sequence \'{seq_name}\'. Incompatible data types.',
            exception.args[0])

    def test_int_float_compatibility(self):
        run = Run(system_tracking_interval=None)

        # float first
        run.track(1., name='float numbers', context={})
        run.track(1, name='float numbers', context={})
        run.track(1., name='float numbers', context={})

        # int first
        run.track(1, name='int numbers', context={})
        run.track(1., name='int numbers', context={})
        run.track(1, name='int numbers', context={})



import logging
import subprocess

logger = logging.getLogger(__name__)


def get_installed_packages():
    import pkg_resources
    packages = {i.key: i.version for i in pkg_resources.working_set}

    return packages


def get_environment_variables():
    from os import environ

    env_mask = ('secret', 'key', 'token', 'password')
    env_vars = {
        k: v for k, v in environ.items() if next(
            (m for m in env_mask if m in k.lower()), None
        ) is None
    }

    return env_vars


def get_git_info():
    git_info = {}
    try:
        r = subprocess.run(['git', 'rev-parse', '--is-inside-work-tree'],
                           stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT,
                           check=True)
    except (subprocess.CalledProcessError, FileNotFoundError):
        # not a git repo
        return git_info
    else:
        output = r.stdout.decode('utf-8').strip().lower()
        if output != 'true':
            # malformed result
            return git_info

    cmds = {
        'branch': ('git', 'rev-parse', '--abbrev-ref', 'HEAD'),
        'remote_origin_url': ('git', 'config', '--get', 'remote.origin.url'),
        'commit': ('git', 'log', '--pretty=format:%h/%ad/%an', '--date=iso-strict', '-1'),
    }
    results = {}
    for key, cmd in cmds.items():
        try:
            r = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, check=True)
        except subprocess.CalledProcessError:
            continue
        else:
            output = r.stdout.decode('utf-8').strip()
            results[key] = output

    try:
        commit_hash, commit_timestamp, commit_author = results.get('commit').split('/')
    except ValueError:
        commit_hash = commit_timestamp = commit_author = None

    git_info.update({
        'branch': results.get('branch'),
        'remote_origin_url': results.get('remote_origin_url'),
        'commit': {
            'hash': commit_hash,
            'timestamp': commit_timestamp,
            'author': commit_author
        }
    })

    return git_info



class TrackerKerasCallbackMetricsEpochEndMixin(object):
    def on_epoch_end(self, epoch, logs=None):
        # Log metrics
        self._log_epoch_metrics(epoch, logs)

    def _get_learning_rate(self):
        lr_schedule = getattr(self.model.optimizer, 'lr', None)
        try:
            return lr_schedule(self.model.optimizer.iterations)
        except Exception:
            return None

    def _log_epoch_metrics(self, epoch, logs):
        if not logs:
            return

        track_func = self._run.track

        train_logs = {k: v for k, v in logs.items() if
                      not k.startswith('val_')}
        for name, value in train_logs.items():
            track_func(value, name=name, epoch=epoch, context={'subset': 'train'})

        val_logs = {k: v for k, v in logs.items() if
                    k.startswith('val_')}
        for name, value in val_logs.items():
            track_func(value, name=name[4:], epoch=epoch, context={'subset': 'val'})

        lr = self._get_learning_rate()
        if lr is not None:
            track_func(lr, name='lr', epoch=epoch, context={'subset': 'train'})



import importlib
import struct
import json
from typing import Iterator, Tuple, Union

import aim.ext.transport.remote_tracking_pb2 as rpc_messages
from aim.storage.object import CustomObject
from aim.storage.types import BLOB

Message = Union[rpc_messages.ResourceRequest, rpc_messages.ResourceResponse]


def pack_args(args: Iterator[Tuple[bytes, bytes]]) -> bytes:
    result = [struct.pack('I', len(key)) + key + struct.pack('I', len(val)) + val for key, val in args]
    return b''.join(result)


def unpack_args(args: bytes) -> Tuple[bytes, bytes]:
    while args:
        (key_size,), args_tail = struct.unpack('I', args[:4]), args[4:]
        key, args_tail = args_tail[:key_size], args_tail[key_size:]

        (value_size,), args_tail = struct.unpack('I', args_tail[:4]), args_tail[4:]
        value, args_tail = args_tail[:value_size], args_tail[value_size:]
        args = args_tail
        yield key, value


def pack_stream(tree: Iterator[Tuple[bytes, bytes]]) -> bytes:
    # TODO: [MV] check the performance diff of current version vs collecting the whole tree as a chunk
    for key, val in tree:
        if not isinstance(val, BLOB):
            yield struct.pack('I', len(key)) + key + struct.pack('?', False) + struct.pack('I', len(val)) + val
        else:
            val = val.load()
            yield struct.pack('I', len(key)) + key + struct.pack('?', True) + struct.pack('I', len(val)) + val


def unpack_helper(msg: bytes) -> Tuple[bytes, bytes]:
    (key_size,), tail = struct.unpack('I', msg[:4]), msg[4:]
    key, tail = tail[:key_size], tail[key_size:]
    (is_blob,), tail = struct.unpack('?', tail[:1]), tail[1:]
    (value_size,), tail = struct.unpack('I', tail[:4]), tail[4:]
    value, tail = tail[:value_size], tail[value_size:]
    assert len(tail) == 0
    if is_blob:
        yield key, BLOB(data=value)
    else:
        yield key, value


def unpack_bytes(stream: Iterator[bytes]) -> Tuple[bytes, bytes]:
    for msg in stream:
        yield from unpack_helper(msg)


def unpack_stream(stream: Iterator[Message]) -> Tuple[bytes, bytes]:
    for msg in stream:
        if msg.WhichOneof('instruction') == 'header':
            # can be header in case of exceptions on server side
            assert msg.header.status == rpc_messages.ResponseHeader.Status.ERROR
            raise_exception(msg.header.exception)
            return

        assert msg.WhichOneof('instruction') == 'message'
        msg = msg.message

        yield from unpack_helper(msg)


def raise_exception(grpc_exception):
    assert grpc_exception is not None
    module = importlib.import_module(grpc_exception.module_name)
    exception = getattr(module, grpc_exception.class_name)
    args = json.loads(grpc_exception.args or [])
    raise exception(*args) if args else exception()


def build_exception(exception: Exception):
    return rpc_messages.ExceptionResponse(
        module_name=exception.__class__.__module__,
        class_name=exception.__class__.__name__,
        args=json.dumps(exception.args),
    )


class UnauthorizedRequestError(RuntimeError):
    def __init__(self, handler, *args, **kwargs):
        self.handler = handler


@CustomObject.alias('aim.resource')
class ResourceObject(CustomObject):
    AIM_NAME = 'aim.resource'

    def __init__(self, handler):
        self.storage['handler'] = handler



import logging

import click
import uuid

from click import core
from collections import OrderedDict
from typing import Optional, Mapping

from aim.ext.notifier import get_config
from aim.ext.notifier.utils import has_watcher_config, set_default_config
from aim.sdk.run_status_watcher import RunStatusWatcher
from aim.sdk.repo import Repo

core._verify_python3_env = lambda: None
DEFAULT_MESSAGE_TEMPLATE = "❗️ Something wrong with Run '{run.hash}'. Please check. ❗️"
MESSAGE_PROMPT = "Stuck Runs notification message"


class OrderedGroup(click.Group):
    def __init__(self, name: Optional[str] = None, commands: Optional[Mapping[str, click.Command]] = None, **kwargs):
        super(OrderedGroup, self).__init__(name, commands, **kwargs)
        #: the registered subcommands by their exported names.
        self.commands = commands or OrderedDict()

    def list_commands(self, ctx: click.Context) -> Mapping[str, click.Command]:
        return self.commands


def check_configuration(ctx: click.Context, repo: Repo) -> bool:
    if not has_watcher_config(repo.path):
        click.echo(f'Repo \'{repo.path}\' has no configured notifiers.')
        if click.confirm('Would you like to configure notifiers?', default=True):
            ctx.invoke(add_config)
        elif click.confirm('Would you like to use default configuration?'):
            set_default_config(repo.path)
        else:
            return False
    return True


def dump_notifier_config(cfg: dict):
    lines = ['', '', f'Type: {cfg["type"]}']
    for arg_name, value in cfg['arguments'].items():
        printable_name = arg_name.replace('_', ' ').capitalize()
        lines.append(f'{printable_name}: {value}')
    lines.append('--------')
    click.echo('\n'.join(lines))


@click.group()
@click.option('--repo',
              required=False,
              help='Aim Repo to check Run statuses.',
              type=click.Path(exists=True,
                              file_okay=False,
                              dir_okay=True,
                              writable=True))
@click.pass_context
def cli_entry_point(ctx, repo):
    """Service for detecting and reporting training Run failures."""
    repo_path = repo or Repo.default_repo_path()
    repo = Repo.from_path(repo_path)

    ctx.ensure_object(dict)
    ctx.obj['repo'] = repo
    ctx.obj['config'] = get_config(repo.path)


@cli_entry_point.command(name='start')
@click.pass_context
def start_watcher(ctx):
    """Start watcher service to monitor and report stuck/failed Runs."""
    repo = ctx.obj['repo']
    if check_configuration(ctx, repo):
        watcher = RunStatusWatcher(repo)
        click.secho(f'Starting Aim watcher for repo \'{repo.path}\'...', fg='yellow')
        click.echo('Press Ctrl+C to exit')
        watcher.start_watcher()


@cli_entry_point.group(cls=OrderedGroup, name='notifiers')
@click.pass_context
def config_notifiers(ctx):
    """Configure how notifications should be received."""
    pass


@click.command(name='dump', hidden=True)
@click.pass_context
def dump_config(ctx):
    """Dump notifier configuration file."""
    cfg = ctx.obj['config']
    if not cfg.exists():
        repo = ctx.obj['repo']
        click.echo(f'Cannot find notifier configuration for Repo \'{repo.path}\'.')
        return

    click.echo(cfg.dump())


@click.command(name='list')
@click.pass_context
def list_config(ctx):
    """List available notifiers."""
    cfg = ctx.obj['config']
    if not cfg.exists():
        repo = ctx.obj['repo']
        click.echo(f'Cannot find notifier configuration for Repo \'{repo.path}\'.')
        return

    click.echo("{:<40} {:<10} {:<10}".format('NOTIFIER ID', 'TYPE', 'STATUS'))
    for notifier in cfg.notifiers.values():
        click.echo("{:<40} {:<10} {:<10}".format(notifier['id'], notifier['type'], notifier['status']))


@click.command(name='get-log-level')
@click.pass_context
def get_log_level(ctx):
    """Get Log Notifications level."""
    cfg = ctx.obj['config']
    if not cfg.exists():
        repo = ctx.obj['repo']
        click.echo(f'Cannot find notifier configuration for Repo \'{repo.path}\'.')
        return

    click.echo(f'Log level: {logging.getLevelName(cfg.log_level)}')


def get_level_names():
    available_levels = (logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG)
    return tuple(map(logging.getLevelName, available_levels))


@click.command(name='set-log-level')
@click.argument('level', required=True, type=click.Choice(get_level_names()))
@click.pass_context
def set_log_level(ctx, level):
    """Set Log Notifications level to <level>."""
    cfg = ctx.obj['config']
    if not cfg.exists():
        repo = ctx.obj['repo']
        click.echo(f'Cannot find notifier configuration for Repo \'{repo.path}\'.')
        return

    cfg.log_level = getattr(logging, level)
    cfg.save()


@click.group(name='add', invoke_without_command=True)
@click.pass_context
def add_config(ctx):
    """Add a new notifier configuration (slack, workplace, etc.)."""
    if ctx.invoked_subcommand is None:
        add_new = True
        while add_new:
            sub_commands = ctx.command.list_commands(ctx)
            choice = click.prompt('Select notifier type to add:', show_choices=True, type=click.Choice(sub_commands))
            sub_cmd = ctx.command.get_command(ctx, choice)
            extra_args = {}
            for param in sub_cmd.params:
                if isinstance(param, click.Option) and param.prompt:
                    extra_args[param.name] = param.prompt_for_value(ctx)
            ctx.invoke(sub_cmd, **extra_args)
            add_new = click.confirm('Would you like to add another notifier?')


@click.command(name='remove')
@click.argument('notifier-id', required=True, type=str)
@click.pass_context
def remove_config(ctx, notifier_id):
    """Remove notifier configuration from the list."""
    cfg = ctx.obj['config']
    if not cfg.exists():
        repo = ctx.obj['repo']
        click.echo(f'Cannot find notifier configuration for Repo \'{repo.path}\'.')
        return

    if notifier_id in cfg.notifiers:
        dump_notifier_config(cfg.get(notifier_id))
        click.confirm('Remove notifier configuration above?', abort=True)
        cfg.remove(notifier_id)
        cfg.save()
    else:
        click.echo(f'No notifier with id {notifier_id} found.')


@click.command(name='enable')
@click.argument('notifier-id', required=True, type=str)
@click.pass_context
def enable_config(ctx, notifier_id):
    """Start receiving notifications from given notifier."""
    cfg = ctx.obj['config']
    if not cfg.exists():
        repo = ctx.obj['repo']
        click.echo(f'Cannot find notifier configuration for Repo \'{repo.path}\'.')
        return

    if notifier_id in cfg.notifiers:
        cfg.enable(notifier_id)
        cfg.save()
    else:
        click.echo(f'No notifier with id {notifier_id} found.')


@click.command(name='disable')
@click.argument('notifier-id', required=True, type=str)
@click.pass_context
def disable_config(ctx, notifier_id):
    """Stop receiving notifications from given notifier."""
    cfg = ctx.obj['config']
    if not cfg.exists():
        repo = ctx.obj['repo']
        click.echo(f'Cannot find notifier configuration for Repo \'{repo.path}\'.')
        return

    if notifier_id in cfg.notifiers:
        cfg.disable(notifier_id)
        cfg.save()
    else:
        click.echo(f'No notifier with id {notifier_id} found.')


@add_config.command(name='workplace')
@click.option('--group-id', prompt=True, required=True, type=int)
@click.option('--access-token', prompt=True, required=True, type=str)
@click.option('--message', prompt=MESSAGE_PROMPT, required=False, type=str,
              default=DEFAULT_MESSAGE_TEMPLATE, show_default=True)
@click.pass_context
def workplace_config(ctx, group_id, access_token, message):
    cfg = ctx.obj['config']
    new_cfg = {
        'id': str(uuid.uuid4()),
        'type': 'workplace',
        'arguments': {
            'group_id': group_id,
            'access_token': access_token,
            'message': message,
        }
    }
    dump_notifier_config(new_cfg)
    click.confirm('Save notifier configuration above?', default=True, abort=True)
    cfg.add(new_cfg)
    cfg.save()


@add_config.command(name='slack')
@click.option('--webhook-url', prompt=True, required=True, type=str)
@click.option('--message', prompt=MESSAGE_PROMPT, required=False, type=str,
              default=DEFAULT_MESSAGE_TEMPLATE, show_default=True)
@click.pass_context
def slack_config(ctx, webhook_url, message):
    cfg = ctx.obj['config']
    new_cfg = {
        'id': str(uuid.uuid4()),
        'type': 'slack',
        'arguments': {
            'url': webhook_url,
            'message': message,
        }
    }
    dump_notifier_config(new_cfg)
    click.confirm('Save notifier configuration above?', default=True, abort=True)
    cfg.add(new_cfg)
    cfg.save()


@add_config.command(name='logger')
@click.option('--message', prompt=MESSAGE_PROMPT, required=False, type=str,
              default=DEFAULT_MESSAGE_TEMPLATE, show_default=True)
@click.pass_context
def logger_config(ctx, message):
    cfg = ctx.obj['config']
    new_cfg = {
        'id': str(uuid.uuid4()),
        'type': 'logger',
        'arguments': {
            'message': message,
        }
    }
    dump_notifier_config(new_cfg)
    click.confirm('Save notifier configuration above?', default=True, abort=True)
    cfg.add(new_cfg)
    cfg.save()


config_notifiers.add_command(add_config)
config_notifiers.add_command(list_config)
config_notifiers.add_command(remove_config)
config_notifiers.add_command(disable_config)
config_notifiers.add_command(enable_config)
config_notifiers.add_command(dump_config)
config_notifiers.add_command(get_log_level)
config_notifiers.add_command(set_log_level)



from typing import Dict

from aim.storage.types import CustomObjectBase
from aim.storage.inmemorytreeview import InMemoryTreeView


class CustomObject(CustomObjectBase):
    registry: Dict[str, type] = {}

    def __init__(self):
        pass

    @staticmethod
    def alias(name: str, exist_ok: bool = True):
        def decorator(cls):
            if name in CustomObject.registry and not exist_ok:
                raise ValueError(f'CustomObject `{name}` is already registered')
            CustomObject.registry[name] = cls
            return cls

        return decorator

    @staticmethod
    def by_name(name: str):
        return CustomObject.registry[name]

    @classmethod
    def get_typename(cls):
        # TODO [AT, MV] check AIM_NAME to have python type name format (f.e. "aim.sdk.Image" but not "abc//}?")
        return cls.AIM_NAME

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        obj.storage = kwargs.get('_storage', InMemoryTreeView(container={}))
        return obj

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}

        cls = self.__class__
        # TODO Implement `__deepcopy__` in `TreeView`
        storage = InMemoryTreeView(container=self.storage[...])
        result = cls.__new__(cls, _storage=storage)
        memodict[id(self)] = result
        return result

    def _aim_encode(self):
        # TODO more effective
        return self.AIM_NAME, self.storage[...]

    @classmethod
    def _aim_decode(cls, aim_name, storage):
        custom_cls = cls.by_name(aim_name)
        return cls.__new__(custom_cls, _storage=storage)



import unittest
import os

from aim.sdk.configs import AIM_REPO_NAME
from performance_tests.conftest import TEST_REPO_PATHS


class TestBase(unittest.TestCase):
    @staticmethod
    def assertInRange(new, base, deviation=0.05):
        upper_limit = base * (1 + deviation)
        lower_limit = base * (1 - deviation)
        failure_message = f'execution time {new}  is out of allowed range [{lower_limit},{upper_limit}]'
        assert lower_limit <= new <= upper_limit, failure_message


class SDKTestBase(TestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        os.environ[AIM_REPO_NAME] = TEST_REPO_PATHS['real_life_repo']

    @classmethod
    def tearDownClass(cls) -> None:
        del os.environ[AIM_REPO_NAME]
        super().tearDownClass()


class StorageTestBase(TestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        os.environ[AIM_REPO_NAME] = TEST_REPO_PATHS['generated_repo']

    @classmethod
    def tearDownClass(cls) -> None:
        del os.environ[AIM_REPO_NAME]
        super().tearDownClass()



import logging
from typing import Optional, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from aim.sdk.repo import Repo

logger = logging.getLogger(__name__)


def get_repo(repo: Optional[Union[str, 'Repo']]) -> 'Repo':
    from aim.sdk.repo import Repo, RepoStatus

    if repo is None:
        repo = Repo.default_repo_path()
    if isinstance(repo, str):
        repo_status = Repo.check_repo_status(repo)
        if repo_status == RepoStatus.UPDATE_REQUIRED:
            logger.error(f'Trying to open repository {repo}, which is out of date. '
                         f'Please upgrade repository with the following command: '
                         f'`aim upgrade --repo {repo} 2to3`.')
            raise RuntimeError()
        elif repo_status == RepoStatus.MISSING:
            repo = Repo.from_path(repo, init=True)
        else:
            repo = Repo.from_path(repo)
    assert isinstance(repo, Repo)

    return repo



import logging
import os
import threading
import uuid
import weakref
from collections import defaultdict
from copy import deepcopy
from typing import Tuple

import aim.ext.transport.remote_tracking_pb2 as rpc_messages
import aim.ext.transport.remote_router_pb2 as router_messages
import aim.ext.transport.remote_tracking_pb2_grpc as remote_tracking_pb2_grpc
import aim.ext.transport.remote_router_pb2_grpc as remote_router_pb2_grpc

from aim.ext.transport.message_utils import pack_stream, unpack_stream, raise_exception
from aim.ext.transport.rpc_queue import RpcQueueWithRetry
from aim.ext.transport.heartbeat import RPCHeartbeatSender
from aim.ext.transport.config import (
    AIM_CLIENT_SSL_CERTIFICATES_FILE,
    AIM_RT_MAX_MESSAGE_SIZE,
    AIM_RT_DEFAULT_MAX_MESSAGE_SIZE,
    AIM_CLIENT_QUEUE_MAX_MEMORY,
)
from aim.storage.treeutils import encode_tree, decode_tree


DEFAULT_RETRY_INTERVAL = 0.1  # 100 ms
DEFAULT_RETRY_COUNT = 2

logger = logging.getLogger(__name__)


class Client:
    _thread_local = threading.local()

    # per run queues. based on run's hash
    _queues = defaultdict(lambda: RpcQueueWithRetry(
        'remote_tracker', max_queue_memory=os.getenv(AIM_CLIENT_QUEUE_MAX_MEMORY, 1024 * 1024 * 1024),
        retry_count=DEFAULT_RETRY_COUNT, retry_interval=DEFAULT_RETRY_INTERVAL))

    def __init__(self, remote_path: str):
        # temporary workaround for M1 build

        self._id = str(uuid.uuid4())
        self._remote_path = remote_path

        self._resource_pool = weakref.WeakValueDictionary()

        self._remote_router_channel = None
        self._remote_router_stub = None
        self._remote_channel = None
        self._remote_stub = None
        self._remote_worker_address = None
        self._create_remote_channels()

        self._heartbeat_sender = RPCHeartbeatSender(self)
        self._heartbeat_sender.start()
        self._thread_local.atomic_instructions = None

    def _create_remote_channels(self):
        import grpc

        ssl_certfile = os.getenv(AIM_CLIENT_SSL_CERTIFICATES_FILE)
        msg_max_size = int(os.getenv(AIM_RT_MAX_MESSAGE_SIZE, AIM_RT_DEFAULT_MAX_MESSAGE_SIZE))
        options = [
            ('grpc.max_send_message_length', msg_max_size),
            ('grpc.max_receive_message_length', msg_max_size)
        ]

        # open a channel with router
        if ssl_certfile:
            with open(ssl_certfile, 'rb') as f:
                root_certificates = grpc.ssl_channel_credentials(f.read())
            self._remote_router_channel = grpc.secure_channel(self.remote_path, root_certificates, options=options)
        elif os.getenv('AIM_FORCE_PUBLIC_CERT') == 'true':
            self._remote_router_channel = grpc.secure_channel(self.remote_path, grpc.ssl_channel_credentials(), options=options)
        else:
            self._remote_router_channel = grpc.insecure_channel(self.remote_path, options=options)

        # get the available worker address
        self._remote_worker_address = self._get_worker_address()

        # if we got the same address as the router, there's no need to open new channels
        if self._remote_worker_address == self.remote_path:
            self._remote_channel = self._remote_router_channel
        elif os.getenv('AIM_FORCE_PUBLIC_CERT') == 'true':
            self._remote_channel = grpc.secure_channel(self._remote_worker_address, grpc.ssl_channel_credentials(), options=options)
        else:
            # open a channel with worker for further communication
            if ssl_certfile:
                self._remote_channel = grpc.secure_channel(self._remote_worker_address,
                                                           root_certificates,
                                                           options=options)
            else:
                self._remote_channel = grpc.insecure_channel(self._remote_worker_address, options=options)

        need_auth = True
        auth_info = None
        if os.getenv('AIM_ACCESS_TOKEN'):
            auth_info = 'Bearer {}'.format(os.getenv('AIM_ACCESS_TOKEN'))
        elif os.getenv('AIM_ACCESS_USERNAME') and os.getenv('AIM_ACCESS_PASSWORD'):
            from base64 import b64encode
            auth_info = 'Basic {}'.format(
                    b64encode('{}:{}'.format(os.getenv('AIM_ACCESS_USERNAME'), os.getenv('AIM_ACCESS_PASSWORD')).encode('utf-8')
                    ).decode('utf-8'))
        elif os.getenv('AIM_ACCESS_CREDENTIAL'):
            auth_info = os.getenv('AIM_ACCESS_CREDENTIAL')
        else:
            need_auth = False

        if need_auth:
            from .client_auth_interceptor import client_auth_interceptor
            header_adder_interceptor = client_auth_interceptor(auth_info)
            self._remote_router_channel = grpc.intercept_channel(self._remote_channel,  header_adder_interceptor)
            if self._remote_worker_address == self.remote_path:
                self._remote_channel = self._remote_router_channel
            else:
                self._remote_channel = grpc.intercept_channel(self._remote_channel,  header_adder_interceptor)

        self._remote_router_stub = remote_router_pb2_grpc.RemoteRouterServiceStub(self._remote_router_channel)

        # check client/server version compatibility
        self._check_remote_version_compatibility()
        self._remote_stub = remote_tracking_pb2_grpc.RemoteTrackingServiceStub(self._remote_channel)

    def reinitialize_resource(self, handler):
        # write some request to get a resource on server side with an already given handler
        resource = self._resource_pool[handler]
        self.get_resource_handler(resource, resource.resource_type, handler, resource.init_args)

    def _reinitialize_all_resources(self):
        handlers_list = list(self._resource_pool.keys())
        for handler in handlers_list:
            self.reinitialize_resource(handler)

    def _check_remote_version_compatibility(self):
        from aim.__version__ import __version__ as client_version
        import grpc

        error_message_template = 'The Aim Remote tracking server version ({}) '\
                                 'is not compatible with the Aim client version ({}).'\
                                 'Please upgrade either the Aim Client or the Aim Remote.'

        warning_message_template = 'The Aim Remote tracking server version ({}) ' \
                                   'and the Aim client version ({}) do not match.' \
                                   'Consider upgrading either the client or remote tracking server.'

        try:
            remote_version = self.get_version()
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNIMPLEMENTED:
                remote_version = '<3.15.0'
            else:
                raise

        # server doesn't yet have the `get_version()` method implemented
        if remote_version == '<3.15.0':
            RuntimeError(error_message_template.format(remote_version, client_version))

        # compare versions
        if client_version == remote_version:
            return

        # if the server has a newer version always force to upgrade the client
        if client_version < remote_version:
            raise RuntimeError(error_message_template.format(remote_version, client_version))

        # for other mismatching versions throw a warning for now
        #logger.warning(warning_message_template.format(remote_version, client_version))
        # further incompatibility list will be added manually

    def _get_worker_address(self):
        worker_host = self._remote_path.rsplit(':', maxsplit=1)[0]
        worker_port = self._get_worker_port()
        return f'{worker_host}:{worker_port}'

    def _get_worker_port(self):
        return int(self._remote_path.rsplit(':', maxsplit=1)[1])

    def client_heartbeat(self):
        request = router_messages.HeartbeatRequest(
            client_uri=self.uri,
        )
        response = self._remote_router_stub.client_heartbeat(request, metadata=[('x-client', self.uri)])
        return response

    def reconnect(self):
        request = router_messages.ReconnectRequest(
            client_uri=self.uri
        )
        response = self._remote_router_stub.reconnect(request, metadata=[('x-client', self.uri)])
        if response.status == router_messages.ReconnectResponse.Status.ERROR:
            raise_exception(response.exception)

        self._reinitialize_all_resources()

        return response

    def disconnect(self):
        request = router_messages.DisconnectRequest(
            client_uri=self.uri
        )
        response = self._remote_router_stub.disconnect(request, metadata=[('x-client', self.uri)])

        if response.status == router_messages.DisconnectResponse.Status.ERROR:
            raise_exception(response.exception)

    def get_version(self,):
        request = router_messages.VersionRequest()
        response = self._remote_router_stub.get_version(request, metadata=[('x-client', self.uri)])

        if response.status == router_messages.VersionResponse.Status.ERROR:
            raise_exception(response.exception)
        return response.version

    def get_resource_handler(self, resource, resource_type, handler='', args=()):
        request = rpc_messages.ResourceRequest(
            resource_type=resource_type,
            handler=handler,
            client_uri=self.uri,
            args=args
        )
        response = self.remote.get_resource(request, metadata=[('x-client', self.uri)])
        if response.status == rpc_messages.ResourceResponse.Status.ERROR:
            raise_exception(response.exception)

        self._resource_pool[response.handler] = resource

        return response.handler

    def release_resource(self, resource_handler):
        request = rpc_messages.ReleaseResourceRequest(
            handler=resource_handler,
            client_uri=self.uri
        )
        response = self.remote.release_resource(request, metadata=[('x-client', self.uri)])
        if response.status == rpc_messages.ReleaseResourceResponse.Status.ERROR:
            raise_exception(response.exception)

        del self._resource_pool[resource_handler]

    def run_instruction(self, queue_id, resource, method, args=(), is_write_only=False):
        args = deepcopy(args)

        # self._thread_local can be empty in the 'clean up' phase.
        if getattr(self._thread_local, 'atomic_instructions', None) is not None:
            assert is_write_only
            self._thread_local.atomic_instructions.append((resource, method, args))
            return

        if is_write_only:
            assert queue_id != -1
            self.get_queue(queue_id).register_task(
                self,
                self._run_write_instructions, list(encode_tree([(resource, method, args)])))
            return

        return self._run_read_instructions(queue_id, resource, method, args)

    def _run_read_instructions(self, queue_id, resource, method, args):
        def message_stream_generator():
            header = rpc_messages.InstructionRequest(
                header=rpc_messages.RequestHeader(
                    version='0.1',
                    handler=resource,
                    client_uri=self.uri,
                    method_name=method
                )
            )
            yield header

            stream = pack_stream(encode_tree(args))
            for chunk in stream:
                yield rpc_messages.InstructionRequest(message=chunk)

        if queue_id != -1:
            self.get_queue(queue_id).wait_for_finish()
        resp = self.remote.run_instruction(message_stream_generator(), metadata=[('x-client', self.uri)])
        status_msg = next(resp)

        assert status_msg.WhichOneof('instruction') == 'header'
        if status_msg.header.status == rpc_messages.ResponseHeader.Status.ERROR:
            raise_exception(status_msg.header.exception)
        return decode_tree(unpack_stream(resp))

    def _run_write_instructions(self, instructions: [Tuple[bytes, bytes]]):
        stream = pack_stream(iter(instructions))

        def message_stream_generator():
            for chunk in stream:
                yield rpc_messages.WriteInstructionsRequest(
                    version='0.1',
                    client_uri=self.uri,
                    message=chunk
                )

        response = self.remote.run_write_instructions(message_stream_generator(), metadata=[('x-client', self.uri)])
        if response.status == rpc_messages.WriteInstructionsResponse.Status.ERROR:
            raise_exception(response.exception)

    def start_instructions_batch(self):
        self._thread_local.atomic_instructions = []

    def flush_instructions_batch(self, queue_id):
        if self._thread_local.atomic_instructions is None:
            return

        self.get_queue(queue_id).register_task(
            self,
            self._run_write_instructions, list(encode_tree(self._thread_local.atomic_instructions)))
        self._thread_local.atomic_instructions = None

    @property
    def remote(self):  # access to low-level interface
        return self._remote_stub

    @property
    def uri(self):
        return self._id

    @property
    def remote_path(self):
        return self._remote_path

    def get_queue(self, queue_id):
        return self._queues[queue_id]

    def remove_queue(self, queue_id):
        del self._queues[queue_id]



import time
import queue
import logging
import threading
import weakref

logger = logging.getLogger(__name__)


class RpcQueueWithRetry(object):
    def __init__(self, name, max_queue_memory=0,
                 retry_count=0, retry_interval=0):

        self._client = None

        self.retry_count = retry_count or 1
        self.retry_interval = retry_interval
        self._needs_reconnect = False

        self.max_memory_usage = max_queue_memory
        self.current_memory_usage = 0

        self._shutdown = False
        self._queue = queue.Queue()
        self._name = name

        self._thread = threading.Thread(target=self.worker)
        self._thread.daemon = True
        self._thread.start()

    def register_task(self, client, task_f, *args):
        if not self._client:
            self._client = weakref.ref(client)

        if self._shutdown:
            logger.debug('Cannot register task: rpc task queue is stopped.')
            return

        arg_size = self._calculate_size(args)
        with self._queue.not_full:
            while self.current_memory_usage + arg_size >= self.max_memory_usage:
                self._queue.not_full.wait()

        with self._queue.mutex:
            self.current_memory_usage += arg_size

        self._queue.put((task_f, args))

    def worker(self):
        while True:
            if self._shutdown:
                logger.debug(f'Shutting down worker thread {threading.get_ident()}.')
                break
            task_f, args = self._queue.get()
            if self._try_exec_task(task_f, *args):
                arg_size = self._calculate_size(args)
                with self._queue.mutex:
                    self.current_memory_usage -= arg_size
                # clear the unnecessary references
                task_f, args = None, None
                self._queue.task_done()
            else:
                self._put_front(task_f, args)

    def _try_exec_task(self, task_f, *args):
        # temporary workaround for M1 build
        import grpc

        retry = 0
        while retry < self.retry_count:
            if self._needs_reconnect:
                try:
                    self._client().reconnect()
                    self._needs_reconnect = False
                except Exception:
                    retry += 1
                    time.sleep(self.retry_interval)
                    continue

            try:
                task_f(*args)
                return True
            except grpc.RpcError as e:
                if e.code() != grpc.StatusCode.UNAVAILABLE:
                    raise e
                self._needs_reconnect = True

                retry += 1
                time.sleep(self.retry_interval)
                logger.warning(f'Remote Server is unavailable, please check network connection: {e}.')

        return False

    def _put_front(self, task_f, args):
        with self._queue.not_full:
            self._queue.queue.appendleft((task_f, args))
            self._queue.not_empty.notify()

    def wait_for_finish(self):
        self._queue.join()

    def stop(self):
        pending_task_count = self._queue.qsize()
        if pending_task_count:
            logger.warning(f'Processing {pending_task_count} pending tasks in the rpc queue \'{self._name}\'... '
                           f'Please do not kill the process.')
            self._queue.join()
        logger.debug('No pending tasks left.')
        self._shutdown = True

    @staticmethod
    def _calculate_size(args):
        size = 0
        assert type(args) is tuple
        for arg in args[0]:
            assert type(arg) is tuple
            assert len(arg) == 2
            size += len(arg[0]) + len(arg[1])
        return size



import io
import logging
import os.path

from aim.sdk.num_utils import inst_has_typename
from aim.sdk.objects.io import wavfile
from aim.storage.object import CustomObject
from aim.storage.types import BLOB

logger = logging.getLogger(__name__)


@CustomObject.alias('aim.audio')
class Audio(CustomObject):
    """Audio object used to store audio objects in Aim repository..

    Currently, audio formats are limited to mp3, wav, flac

    Args:
         data: file path, bytes, io.BaseIO or numpy.array (only for WAV)
         format (:obj:`str`): Format of the audio source
         rate (:obj:`int`): Rate of the audio file, for WAV defaults to 22500
         caption (:obj:`str`, optional): Optional audio caption. '' by default.
    """

    AIM_NAME = 'aim.audio'

    # supported audio formats
    UNKNOWN = ''
    MP3 = 'mp3'
    WAV = 'wav'
    FLAC = 'flac'

    audio_formats = (MP3, WAV, FLAC)

    def __init__(self, data, format: str = '', caption: str = '', rate: int = None):
        super().__init__()

        audio_format = format.lower()
        if inst_has_typename(data, ['ndarray.numpy']):
            # Currently, only WAV audio formats are supported for numpy
            audio_format = self.WAV
            if not rate:
                rate = 22500
                logger.info(f'Parameter "rate" is not provided! Using default: {rate}')
            bs = wavfile.write(rate, data)
            data = bs

        # act as a regular file with enforced audio format definition by user side
        if not audio_format:
            raise ValueError('Audio format must be provided.')
        elif audio_format not in self.audio_formats:
            raise ValueError(f'Invalid audio format is provided. Must be one of {self.audio_formats}')

        if isinstance(data, str):
            if not os.path.exists(data) or not os.path.isfile(data):
                raise ValueError('Invalid audio file path')
            with open(data, 'rb') as FS:
                data = FS.read()
        elif isinstance(data, io.BytesIO):
            data = data.read()

        if not isinstance(data, bytes):
            raise TypeError('Content is not a byte-stream object')

        extra = {
            'caption': caption,
            'format': audio_format
        }
        self._prepare(data, **extra)

    def json(self):
        """Dump audio metadata to a dict"""
        return {
            'caption': self.storage['caption'],
            'format': self.storage['format']
        }

    def _prepare(self, data, **extra) -> None:
        assert isinstance(data, bytes)

        for k, v in extra.items():
            self.storage[k] = v
        self.storage['data'] = BLOB(data=data)

    def to_numpy(self):
        """
        This method converts WAV to Numpy array.
        Other audio formats are not supported at this moment.

        Returns: numpy array
        """
        assert self.storage['format'] == self.WAV

        return wavfile.read(self.get())

    def get(self) -> io.BytesIO:
        """
        Reads data from the inner container and writes it to a buffer

        Returns: io.BytesIO
        """
        bs = self.storage.get('data')
        if not bs:
            return io.BytesIO()
        return io.BytesIO(bytes(bs))



from aim.ext.notifier.notifier import Notifier, NotificationSendError
from aim.ext.notifier.notifier_builder import NotifierBuilder
from aim.ext.notifier.config import Config
from aim.ext.notifier.utils import get_config_path


def get_config(base_dir) -> Config:
    config_file = get_config_path(base_dir)
    return Config(config_file)


def get_notifier(base_dir) -> Notifier:
    cfg = get_config(base_dir)
    cfg.load()
    builder = NotifierBuilder(cfg.notifiers)
    return builder.build()


__all__ = ['Notifier', 'NotificationSendError', 'Config', 'get_config', 'get_notifier']



from importlib import import_module
import os
import subprocess

from aim.web.configs import AIM_UI_MOUNTED_REPO_PATH
from aim.sdk.configs import get_aim_repo_name
from aim.sdk.utils import clean_repo_path


class ShellCommandException(Exception):
    pass


def exec_cmd(
    cmd, throw_on_error=True, env=None, stream_output=False, cwd=None, cmd_stdin=None, **kwargs
):
    """
    Runs a command as a child process.
    A convenience wrapper for running a command from a Python script.
    Keyword arguments:
    cmd -- the command to run, as a list of strings
    throw_on_error -- if true, raises an Exception if the exit code of the program is nonzero
    env -- additional environment variables to be defined when running the child process
    cwd -- working directory for child process
    stream_output -- if true, does not capture standard output and error; if false, captures these
      streams and returns them
    cmd_stdin -- if specified, passes the specified string as stdin to the child process.
    Note on the return value: If stream_output is true, then only the exit code is returned. If
    stream_output is false, then a tuple of the exit code, standard output and standard error is
    returned.
    """
    cmd_env = os.environ.copy()
    if env:
        cmd_env.update(env)
    if stream_output:
        child = subprocess.Popen(
            cmd, env=cmd_env, cwd=cwd, universal_newlines=True, stdin=subprocess.PIPE, **kwargs
        )
        child.communicate(cmd_stdin)
        exit_code = child.wait()
        if throw_on_error and exit_code != 0:
            raise ShellCommandException("Non-zero exitcode: %s" % (exit_code))
        return exit_code
    else:
        child = subprocess.Popen(
            cmd,
            env=cmd_env,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=cwd,
            universal_newlines=True,
            **kwargs
        )
        (stdout, stderr) = child.communicate(cmd_stdin)
        exit_code = child.wait()
        if throw_on_error and exit_code != 0:
            raise ShellCommandException(
                "Non-zero exit code: %s\n\nSTDOUT:\n%s\n\nSTDERR:%s" % (exit_code, stdout, stderr)
            )
        return exit_code, stdout, stderr


def get_module(name, required=True):
    try:
        return import_module(name)
    except Exception:
        if required:
            raise ValueError('No module named: \'{}\''.format(name))
        return None


def ls_dir(path):
    """
    List the files in directories
    """
    if not path or not os.path.exists:
        return []

    if os.path.isfile(path):
        return [path]

    ls = []

    for root, _, file_names in os.walk(path):
        for file_name in file_names:
            ls.append(os.path.join(root, file_name))

    return ls


def get_root_path():
    return clean_repo_path(os.getenv(AIM_UI_MOUNTED_REPO_PATH, os.getcwd()))


def get_db_url():
    return 'sqlite:///{}/{}/aim_db'.format(get_root_path(), get_aim_repo_name())



from typing import Optional, List

from aim.sdk.run import Run
from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT

try:
    from kerastuner.engine.tuner_utils import TunerCallback
except ImportError:
    raise RuntimeError(
        'This contrib module requires KerasTuner to be installed. '
        'Please install it with command: \n pip install keras-tuner'
    )


class AimCallback(TunerCallback):
    def __init__(self, tuner=None,
                 repo: Optional[str] = None,
                 experiment: Optional[str] = None,
                 system_tracking_interval: int = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: bool = True):
        self.tuner = tuner
        self._repo_path = repo
        self._experiment_name = experiment
        self._system_tracking_interval = system_tracking_interval
        self._log_system_params = log_system_params

        self._started_trials: List[str] = []
        self.trial = None
        self._run = None

    @property
    def experiment(self) -> Run:
        if self._run is not None:
            return self._run

    def on_epoch_begin(self, epoch, logs=None):
        trial_dict = self.tuner.oracle.ongoing_trials
        tuner_key = next(iter(trial_dict))
        self._current_trial_id = trial_dict[tuner_key].trial_id
        if self._current_trial_id not in self._started_trials:
            if self._repo_path is None and self._experiment_name is None:
                self._run = Run(system_tracking_interval=self._system_tracking_interval,
                                log_system_params=self._log_system_params,)
            else:
                self._run = Run(repo=self._repo_path, experiment=self._experiment_name,
                                system_tracking_interval=self._system_tracking_interval,
                                log_system_params=self._log_system_params,)
            self._run['trial_id'] = self._current_trial_id
            self._started_trials.append(self._current_trial_id)
        trial = self.tuner.oracle.get_trial(self._current_trial_id)
        hparams = trial.hyperparameters.values
        for key in hparams:
            self._run.set(key, hparams[key], strict=False)

    def on_batch_end(self, batch, logs=None):
        if logs:
            for log_name, log_value in logs.items():
                self._run.track(log_value, name=log_name)

    def __del__(self):
        if self._run is not None and self._run.active:
            self._run.close()



from aim.cli.upgrade._legacy_repo.proto.base_pb2 import BaseRecord
from aim.cli.upgrade._legacy_repo.proto.metric_pb2 import MetricRecord


def deserialize_pb_object(data) -> BaseRecord:
    base_pb = BaseRecord()
    base_pb.ParseFromString(data)
    return base_pb


def deserialize_pb(data):
    base_pb = deserialize_pb_object(data)

    metric_pb = MetricRecord()
    metric_pb.ParseFromString(base_pb.artifact)

    return base_pb, metric_pb



import logging

from aim.sdk.num_utils import inst_has_typename
from aim.storage.object import CustomObject
from aim.storage.types import BLOB

logger = logging.getLogger(__name__)


@CustomObject.alias('aim.figure')
class Figure(CustomObject):
    """
    Figure object can be used for storing Plotly or Matplotlib figures into Aim repository.
    Core functionality is based on Plotly.

    Args:
         obj (:obj:): plotly or matplotlib figure object.
    """

    AIM_NAME = 'aim.figure'

    def __init__(self, obj):
        super().__init__()

        if inst_has_typename(obj, ['matplotlib', 'Figure']):
            self._from_matplotlib_figure(obj)
        elif inst_has_typename(obj, ['plotly', 'Figure', 'BaseFigure']):
            self._prepare(obj)
        else:
            raise TypeError('Object is not a Plotly Figure instance')

    def _prepare(self, obj):
        try:
            from plotly.version import __version__ as plotly_version
        except ModuleNotFoundError:
            plotly_version = 'unknown'

        assert hasattr(obj, 'to_json')

        self.storage['source'] = 'plotly'
        self.storage['version'] = plotly_version
        self.storage['format'] = 'raw_json'
        self.storage['data'] = BLOB(data=obj.to_json())

    @property
    def data(self):
        return self.storage['data'].load()

    def _from_matplotlib_figure(self, obj):
        try:
            from plotly.tools import mpl_to_plotly
        except ModuleNotFoundError:
            raise ModuleNotFoundError('Plotly is required to track matplotlib figure.')

        try:
            logger.warning('Tracking a matplotlib object using "aim.Figure" might not behave as expected.'
                           'In such cases, consider tracking with "aim.Image".')
            for ax in obj.axes:
                for collection in ax.collections:
                    if not hasattr(collection, "get_offset_position"):
                        collection.get_offset_position = matplotlib_get_offset_position.__get__(collection)
            plotly_obj = mpl_to_plotly(obj)
        except ValueError as err:
            raise ValueError(f'Failed to convert matplotlib figure to plotly figure: {err}')

        return self._prepare(plotly_obj)

    def json(self):
        """Dump figure metadata to a dict"""
        return {
            'source': self.storage['source'],
            'format': self.storage['format'],
            'version': self.storage['version']
        }

    def to_plotly_figure(self):
        try:
            from plotly.io import from_json
        except ModuleNotFoundError:
            raise ModuleNotFoundError('Could not find plotly in the installed modules.')

        return from_json(self.data)


def matplotlib_get_offset_position(self):
    # self._offset_position is depricated
    # using 'screen' as the default value
    return 'screen'



import asyncio
import numpy as np
import random
import struct
import time

from collections import namedtuple
from itertools import chain
from typing import Iterator, Tuple, Optional, List
from typing import TYPE_CHECKING

from fastapi import HTTPException

from aim.storage.context import Context
from aim.sdk import Run
from aim.sdk.sequences.metric import Metric
from aim.sdk.sequence_collection import SequenceCollection
from aim.storage.query import syntax_error_check
from aim.web.configs import AIM_PROGRESS_REPORT_INTERVAL
from aim.web.api.projects.project import Project
from aim.web.api.runs.pydantic_models import AlignedRunIn, TraceBase
from aim.storage.treeutils import encode_tree

if TYPE_CHECKING:
    from aim.sdk import Repo

IndexRange = namedtuple('IndexRange', ['start', 'stop'])

# added to progress keys to escape buffering due to gzipping responses
PROGRESS_KEY_SUFFIX = ''.join([str(hash(random.random())) for _ in range(2000)])

# Used for async sleep to mimic real async streamer
# reference here
# https://github.com/tiangolo/fastapi/issues/4146
ASYNC_SLEEP_INTERVAL = 0.00001


def get_run_or_404(run_id, repo=None):
    if repo is None:
        repo = get_project_repo()

    run = repo.get_run(run_id)
    if not run:
        raise HTTPException(status_code=404, detail="Run not found.")

    return run


def str_to_range(range_str: str):
    defaults = [None, None]
    slice_values = chain(range_str.strip().split(':'), defaults)

    start, stop, step, *_ = map(lambda x: int(x) if x else None, slice_values)
    return IndexRange(start, stop)


def get_run_params(run: Run, *, skip_system: bool):
    params = run.get(..., {}, resolve_objects=True)
    if skip_system and '__system_params' in params:
        del params['__system_params']
    return params


def get_run_props(run: Run):
    return {
        'name': run.name if run.name else None,
        'description': run.description if run.description else None,
        'experiment': {
            'id': run.props.experiment_obj.uuid,
            'name': run.props.experiment_obj.name,
            'description': run.props.experiment_obj.description,
        } if run.props.experiment_obj else None,
        'tags': [{'id': tag.uuid,
                  'name': tag.name,
                  'color': tag.color,
                  'description': tag.description}
                 for tag in run.props.tags_obj],
        'archived': run.archived if run.archived else False,
        'creation_time': run.creation_time,
        'end_time': run.end_time,
        'active': run.active
    }


def numpy_to_encodable(array: np.ndarray) -> Optional[dict]:
    encoded_numpy = {
        'type': 'numpy',
        'shape': array.shape[0],
        'dtype': 'float64',  # hardcoded for now
    }

    if array.dtype == 'float64':
        encoded_numpy['blob'] = array.tobytes()
    elif array.dtype == 'object':
        return None
    else:
        encoded_numpy['blob'] = array.astype('float64').tobytes()
    return encoded_numpy


def collect_x_axis_data(x_trace: Metric, iters: np.ndarray) -> Tuple[Optional[dict], Optional[dict]]:
    if not x_trace:
        return None, None

    x_axis_values = []
    x_axis_iters = []
    step_index_fn = x_trace.data.step_hash
    for idx in iters:
        try:
            x_val = x_trace.values[step_index_fn(idx.item())]
        except KeyError:
            x_val = None
        if x_val:
            x_axis_iters.append(step_index_fn(idx.item()))
            x_axis_values.append(x_val)

    if not x_axis_iters:
        return None, None

    return (
        numpy_to_encodable(np.array(x_axis_iters, dtype='float64')),
        numpy_to_encodable(np.array(x_axis_values, dtype='float64'))
    )


def collect_streamable_data(encoded_tree: Iterator[Tuple[bytes, bytes]]) -> bytes:
    result = [struct.pack('I', len(key)) + key + struct.pack('I', len(val)) + val for key, val in encoded_tree]
    return b''.join(result)


async def custom_aligned_metrics_streamer(requested_runs: List[AlignedRunIn], x_axis: str, repo: 'Repo') -> bytes:
    try:
        for run_data in requested_runs:
            await asyncio.sleep(ASYNC_SLEEP_INTERVAL)
            run_hash = run_data.run_id
            requested_traces = run_data.traces
            run = Run(run_hash, repo=repo, read_only=True)

            traces_list = []
            for trace_data in requested_traces:
                context = Context(trace_data.context)
                trace = run.get_metric(name=trace_data.name,
                                       context=context)
                x_axis_trace = run.get_metric(name=x_axis,
                                              context=context)
                if not (trace and x_axis_trace):
                    continue

                iters = np.array(trace.data.sample(trace_data.slice[-1]).indices_list())
                x_axis_iters, x_axis_values = collect_x_axis_data(x_axis_trace, iters)
                traces_list.append({
                    'name': trace.name,
                    'context': trace.context.to_dict(),
                    'x_axis_values': x_axis_values,
                    'x_axis_iters': x_axis_iters,
                })
            run_dict = {
                run_hash: traces_list
            }
            encoded_tree = encode_tree(run_dict)
            yield collect_streamable_data(encoded_tree)
    except asyncio.CancelledError:
        pass


async def metric_search_result_streamer(traces: SequenceCollection,
                                        skip_system: bool,
                                        steps_num: int,
                                        x_axis: Optional[str] = None,
                                        report_progress: Optional[bool] = True) -> bytes:
    try:
        last_reported_progress_time = time.time()
        progress = None
        progress_reports_sent = 0
        for run_trace_collection, progress in traces.iter_runs():
            await asyncio.sleep(ASYNC_SLEEP_INTERVAL)
            if report_progress and time.time() - last_reported_progress_time > AIM_PROGRESS_REPORT_INTERVAL:
                yield collect_streamable_data(encode_tree(
                    {f'progress_{progress_reports_sent}_{PROGRESS_KEY_SUFFIX}': progress}
                ))
                progress_reports_sent += 1
                last_reported_progress_time = time.time()

            run = None
            traces_list = []
            for trace in run_trace_collection.iter():
                if not run:
                    run = run_trace_collection.run
                iters, (values, epochs, timestamps) = trace.data.sample(steps_num).numpy()

                x_axis_trace = run.get_metric(x_axis, trace.context) if x_axis else None
                x_axis_iters, x_axis_values = collect_x_axis_data(x_axis_trace, iters)

                traces_list.append({
                    'name': trace.name,
                    'context': trace.context.to_dict(),
                    'slice': [0, 0, steps_num],  # TODO [AT] change once UI is ready
                    'values': numpy_to_encodable(values),
                    'iters': numpy_to_encodable(iters),
                    'epochs': numpy_to_encodable(epochs),
                    'timestamps': numpy_to_encodable(timestamps),
                    'x_axis_values': x_axis_values,
                    'x_axis_iters': x_axis_iters,
                })

            if run:
                run_dict = {
                    run.hash: {
                        'params': get_run_params(run, skip_system=skip_system),
                        'traces': traces_list,
                        'props': get_run_props(run)
                    }
                }

                encoded_tree = encode_tree(run_dict)
                yield collect_streamable_data(encoded_tree)
                if report_progress:
                    yield collect_streamable_data(encode_tree({f'progress_{progress_reports_sent}': progress}))
                    progress_reports_sent += 1
                    last_reported_progress_time = time.time()

        if report_progress and progress:
            yield collect_streamable_data(encode_tree({f'progress_{progress_reports_sent}': progress}))
    except asyncio.CancelledError:
        pass


async def run_search_result_streamer(runs: SequenceCollection,
                                     limit: int,
                                     skip_system: bool,
                                     report_progress: Optional[bool] = True,
                                     exclude_params: Optional[bool] = False,
                                     exclude_traces: Optional[bool] = False) -> bytes:
    try:
        run_count = 0
        last_reported_progress_time = time.time()
        progress = None
        progress_reports_sent = 0
        for run_trace_collection, progress in runs.iter_runs():
            await asyncio.sleep(ASYNC_SLEEP_INTERVAL)
            # if no progress was reported for a long interval, report progress
            if report_progress and time.time() - last_reported_progress_time > AIM_PROGRESS_REPORT_INTERVAL:
                yield collect_streamable_data(encode_tree(
                    {f'progress_{progress_reports_sent}_{PROGRESS_KEY_SUFFIX}': progress}
                ))
                progress_reports_sent += 1
                last_reported_progress_time = time.time()
            if not run_trace_collection:
                continue
            run = run_trace_collection.run
            run_dict = {
                run.hash: {
                    'props': get_run_props(run)
                }
            }
            if not exclude_params:
                run_dict[run.hash]['params'] = get_run_params(run, skip_system=skip_system)
            if not exclude_traces:
                run_dict[run.hash]['traces'] = run.collect_sequence_info(sequence_types='metric')

            encoded_tree = encode_tree(run_dict)
            yield collect_streamable_data(encoded_tree)
            if report_progress:
                yield collect_streamable_data(encode_tree({f'progress_{progress_reports_sent}': progress}))
                progress_reports_sent += 1
                last_reported_progress_time = time.time()
            run_count += 1
            if limit and run_count >= limit:
                break

        if report_progress and progress:
            yield collect_streamable_data(encode_tree({f'progress_{progress_reports_sent}': progress}))
    except asyncio.CancelledError:
        pass


async def run_active_result_streamer(repo: 'Repo', report_progress: Optional[bool] = True):
    try:
        active_run_hashes = repo.list_active_runs()

        active_runs_count = len(active_run_hashes)
        progress_reports_sent = 0

        for run_hash in active_run_hashes:
            await asyncio.sleep(ASYNC_SLEEP_INTERVAL)

            run = Run(run_hash, repo=repo, read_only=True)
            if run.active:
                run_dict = {
                    run.hash: {
                        'props': get_run_props(run),
                        'traces': run.collect_sequence_info(sequence_types='metric')
                    }
                }

                encoded_tree = encode_tree(run_dict)
                yield collect_streamable_data(encoded_tree)

            if report_progress:
                yield collect_streamable_data(encode_tree(
                    {f'progress_{progress_reports_sent}': (progress_reports_sent + 1, active_runs_count)}
                ))
                progress_reports_sent += 1
    except asyncio.CancelledError:
        pass


def collect_requested_metric_traces(run: Run, requested_traces: List[TraceBase], steps_num: int = 200) -> List[dict]:
    processed_traces_list = []
    for requested_trace in requested_traces:
        metric_name = requested_trace.name
        context = Context(requested_trace.context)
        trace = run.get_metric(name=metric_name, context=context)
        if not trace:
            continue

        iters, (values,) = trace.data.view('val').sample(steps_num).items_list()
        values = list(map(lambda x: x if float('-inf') < x < float('inf') and x == x else None, values))

        processed_traces_list.append({
            'name': trace.name,
            'context': trace.context.to_dict(),
            'values': values,
            'iters': iters,
        })

    return processed_traces_list


async def run_logs_streamer(run: Run, record_range: str) -> bytes:
    logs = run.get_terminal_logs()

    if not logs:
        return

    record_range = checked_range(record_range)
    start = record_range.start
    stop = record_range.stop

    # range stop is missing
    if record_range.stop is None:
        stop = logs.last_step() + 1

    # range start is missing
    if record_range.start is None:
        start = 0

    # range is missing completely
    if record_range.start is None and record_range.stop is None:
        start = max(logs.last_step() - 200, 0)

    try:
        steps_vals = logs.data.view('val').range(start, stop)
        for step, (val,) in steps_vals:
            await asyncio.sleep(ASYNC_SLEEP_INTERVAL)
            encoded_tree = encode_tree({step: val.data})
            yield collect_streamable_data(encoded_tree)
    except asyncio.CancelledError:
        pass


def get_project():
    project = Project()
    if not project.exists():
        raise HTTPException(status_code=404)
    return project


def get_project_repo():
    project = get_project()
    return project.repo


def checked_query(q: str):
    query = q.strip()
    try:
        syntax_error_check(query)
    except SyntaxError as se:
        raise HTTPException(status_code=400, detail={
            'message': 'SyntaxError',
            'detail': {
                'statement': se.text,
                'line': se.lineno,
                'offset': se.offset,
                'end_offset': getattr(se, 'end_offset', 0)
            }
        })
    return query


def checked_range(range_: str = ''):
    try:
        range_ = str_to_range(range_)
    except ValueError:
        raise HTTPException(status_code=400, detail='Invalid range format')
    return range_



from typing import Optional

from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT
from aim.sdk.run import Run

try:
    from xgboost.callback import TrainingCallback
except ImportError:
    raise RuntimeError(
        'This contrib module requires XGBoost to be installed. '
        'Please install it with command: \n pip install xgboost'
    )


class AimCallback(TrainingCallback):

    def __init__(self, repo: Optional[str] = None,
                 experiment: Optional[str] = None,
                 system_tracking_interval: Optional[int]
                 = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: Optional[int] = True):
        super().__init__()
        self._repo_path = repo
        self._experiment = experiment
        self._system_tracking_interval = system_tracking_interval
        self._log_system_params = log_system_params
        self._run = None
        self._run_hash = None

    @property
    def experiment(self) -> Run:
        if not self._run:
            self.setup()
        return self._run

    def setup(self):
        if self._run:
            return
        if self._run_hash:
            self._run = Run(
                self._run_hash,
                repo=self._repo_path,
                system_tracking_interval=self._system_tracking_interval,
            )
        else:
            self._run = Run(
                repo=self._repo_path,
                experiment=self._experiment,
                system_tracking_interval=self._system_tracking_interval,
                log_system_params=self._log_system_params
            )
            self._run_hash = self._run.hash

    def before_training(self, model):
        self.setup()
        return model

    def after_iteration(self, model, epoch: int, evals_log) -> bool:
        if not evals_log:
            return False

        for data, metric in evals_log.items():
            for metric_name, log in metric.items():
                stdv: Optional[float] = None
                if isinstance(log[-1], tuple):
                    score = log[-1][0]
                    stdv = log[-1][1]
                else:
                    score = log[-1]

                self._run.track(score, step=0, name=metric_name, context={'stdv': False})
                if stdv is not None:
                    self._run.track(score, step=0, name=metric_name, context={'stdv': True})

        return False

    def after_training(self, model):
        self.close()
        return model

    def close(self):
        if self._run:
            self._run.close()
            del self._run
            self._run = None

    def __del__(self):
        self.close()



import time

from inspect import getframeinfo, currentframe
from typing import Optional, Union, Tuple

from aim.storage.object import CustomObject
from aim.storage.hashing import hash_auto

from aim.sdk.sequence import Sequence


@CustomObject.alias('aim.log_record')
class LogRecord(CustomObject):
    AIM_NAME = 'aim.log_record'

    def __init__(self,
                 message: str,
                 level: int,
                 timestamp: Optional[float] = None,
                 logger_info: Optional[Tuple[str, int]] = None,
                 **kwargs):
        super().__init__()
        self.storage['message'] = message
        self.storage['log_level'] = level
        self.storage['timestamp'] = timestamp or time.time()
        for k, v in kwargs:
            self.storage['extra_args'].set(k, v, strict=False)

        if not logger_info:
            frame_info = getframeinfo(currentframe().f_back)
            logger_info = (frame_info.filename, frame_info.lineno)
        self.storage['__logger_info'] = logger_info

    @property
    def message(self):
        return self.storage['message']

    @property
    def level(self):
        return self.storage['log_level']

    def __hash__(self):
        return hash_auto((self.storage['__logger_info'], self.storage['log_level']))


class LogRecords(Sequence):
    """Class representing series of LogLine objects."""

    @classmethod
    def allowed_dtypes(cls) -> Union[str, Tuple[str, ...]]:
        text_typename = LogRecord.get_typename()
        return text_typename

    @classmethod
    def sequence_name(cls) -> str:
        return 'log_records'



import threading
import queue
import atexit
import logging

logger = logging.getLogger(__name__)


class TaskQueue(object):
    def __init__(self, name, num_workers=1, max_backlog=0):
        self.name = name
        self.max_backlog = max_backlog
        self.num_workers = num_workers

        self._queue = queue.Queue(maxsize=max_backlog)

        self._threads = []
        self._shutdown = False
        self._stopped = False
        atexit.register(self.stop_workers)

        for thread_num in range(self.num_workers):
            thread = threading.Thread(target=self.worker)
            thread.daemon = True
            self._threads.append(thread)
            thread.start()

    def register_task(self, task_func, *args, **kwargs):
        warn_queue_full = False
        if self._stopped:
            logger.debug('Cannot register task: task queue is stopped.')
        else:
            backlog_size = self._queue.qsize()
            if backlog_size > self.max_backlog * 0.8:  # queue is 80% full
                warn_queue_full = True
            self._queue.put((task_func, args, kwargs))
        return warn_queue_full

    def worker(self):
        while True:
            if self._shutdown:
                logger.debug(f'Shutting down worker thread {threading.get_ident()}.')
                break
            task_f, args, kwargs = self._queue.get()
            task_f(*args, **kwargs)
            # clear the unnecessary references to Run objects
            task_f, args, kwargs = None, None, None
            self._queue.task_done()

    def stop_workers(self):
        if self._stopped:
            return

        self._stopped = True
        pending_task_count = self._queue.qsize()
        if pending_task_count:
            logger.warning(f'Processing {pending_task_count} pending tasks in queue \'{self.name}\'... '
                           f'Please do not kill the process.')
            self._queue.join()
        self._shutdown = True
        logger.debug('No pending tasks left.')

    def __del__(self):
        self.stop_workers()



import inspect

from typing import Callable, Set


def handles_events(method: Callable) -> bool:
    return hasattr(method, '_handled_events_set_')


def get_handler_event_names(method: Callable) -> Set[str]:
    return getattr(method, '_handled_events_set_', set())


def add_handler_event_name(method: Callable, event_name: str):
    event_names = getattr(method, '_handled_events_set_', set())
    event_names.add(event_name)
    setattr(method, '_handled_events_set_', event_names)


def check_handler_method_params(method: Callable):
    # TODO better error handling. actionable messages
    signature = inspect.signature(method)
    first, *params, last = list(signature.parameters.items())

    first_param_name, _ = first
    assert first_param_name == 'self'
    for _, param in params:
        assert param.kind not in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.VAR_POSITIONAL)
    _, last_param = last
    assert last_param.kind == inspect.Parameter.VAR_KEYWORD


def predefine_event(func):
    forbidden_param_kinds = (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.VAR_POSITIONAL)
    signature = inspect.signature(func)
    for param in signature.parameters.values():
        assert param.kind not in forbidden_param_kinds

    def handle_event(self, **kwargs):
        # TODO allow calls w/ positional arguments?
        self.trigger(func.__name__, **kwargs)

    return handle_event



import datetime
import uuid
import psutil
import os
import time
import logging
import weakref

from typing import Union
from enum import Enum
from pathlib import Path
from dataclasses import dataclass, field
from dateutil.relativedelta import relativedelta
from filelock import UnixFileLock, SoftFileLock, Timeout

from aim.sdk.errors import RunLockingError
from aim.storage.locking import RunLock

logger = logging.getLogger(__name__)


class LockingVersion(Enum):
    LEGACY = 0
    NEW = 1


class LockType(Enum):
    SOFT_LOCK = 0
    UNIX_LOCK = 1


@dataclass(frozen=True)
class LockInfo:
    run_hash: str = field()
    locked: bool = field(default=False)
    version: LockingVersion = field(default=LockingVersion.NEW)
    type: LockType = field(default=LockType.SOFT_LOCK)
    created_at: datetime.datetime = field(default=None)

    def age(self, to=None) -> str:
        if self.created_at is None:
            return 'N/A'
        to = to or datetime.datetime.now()
        delta = relativedelta(to, self.created_at)
        date_attrs = ('years', 'months', 'days', 'hours', 'minutes', 'seconds')
        for attr in date_attrs:
            if getattr(delta, attr) > 1:
                return f'~{getattr(delta, attr)} {attr}'
            elif getattr(delta, attr) == 1:
                return f'~{getattr(delta, attr)} {attr[:-1]}'


class SFRunLock(RunLock):
    def __init__(self, lock_manager: 'LockManager', run_hash: str, path: Path, timeout: int):
        self.run_hash = run_hash
        self._lock_manager = weakref.ref(lock_manager)
        self._sf_lock = SoftFileLock(path, timeout=timeout)

    def lock(self, force: bool = False):
        self._lock_manager().lock(self.run_hash, self._sf_lock, force=force)

    def release(self, force: bool = False) -> None:
        self._sf_lock.release(force=force)


class LockManager(object):
    machine_id = uuid.getnode()
    pid = os.getpid()

    def __init__(self, repo_path: Union[str, Path]):
        self.repo_path = Path(repo_path)
        self.locks_path = self.repo_path / 'locks'
        self.locks_path.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def softlock_fname(name: str) -> str:
        return f'{name}.softlock'

    def get_run_lock_info(self, run_hash: str) -> LockInfo:
        # check locks created prior to 3.15 version
        locked = False
        created_at = None
        lock_version = None
        lock_type = None

        run_lock_path = self.locks_path / self.softlock_fname(run_hash)
        if run_lock_path.exists():
            locked = True
            created_at = datetime.datetime.fromtimestamp(run_lock_path.stat().st_mtime)
            lock_version = LockingVersion.NEW
            lock_type = LockType.SOFT_LOCK
        else:
            # consider Run as locked if one of it's containers is locked
            for container_dir in ('meta', 'seqs'):
                lock_dir = self.repo_path / container_dir / 'locks'
                lock_path = lock_dir / run_hash
                soft_lock_path = lock_dir / self.softlock_fname(run_hash)
                if lock_path.exists():
                    try:
                        lock = UnixFileLock(lock_path, timeout=0)
                        with lock.acquire():
                            pass
                    except Timeout:
                        locked = True
                        lock_version = LockingVersion.LEGACY
                        lock_type = LockType.UNIX_LOCK
                elif soft_lock_path.exists():
                    locked = True
                    created_at = datetime.datetime.fromtimestamp(soft_lock_path.stat().st_mtime)
                    lock_version = LockingVersion.LEGACY
                    lock_type = LockType.SOFT_LOCK

        return LockInfo(run_hash=run_hash, locked=locked, created_at=created_at, version=lock_version, type=lock_type)

    def get_run_lock(self, run_hash: str, timeout: int = 10) -> RunLock:
        lock_path = self.locks_path / self.softlock_fname(run_hash)
        return SFRunLock(self, run_hash, lock_path, timeout=timeout)

    def lock(self, run_hash: str, run_lock: SoftFileLock, force: bool = False):
        lock_path = Path(run_lock.lock_file)

        if force:
            logger.warning(f'Force-releasing locks for Run \'{run_hash}\'. Data corruption may occur if there is '
                           f'active process writing to Run \'{run_hash}\'.')
            self.release_locks(run_hash, force=True)
        elif not self.release_locks(run_hash, force=False):
            raise RunLockingError(f'Cannot acquire lock for Run \'{run_hash}\'. '
                                  f'Make sure no process uses Run \'{run_hash}\' and close it via Aim CLI:\n'
                                  f'`aim runs close --force {run_hash}`')
        run_lock.acquire()
        with open(lock_path, 'w') as lock_metadata_fh:
            lock_metadata_fh.write(f'{self.machine_id}-{self.pid}-{time.time()}')

    def release_locks(self, run_hash: str, force: bool) -> bool:
        success = True
        lock_path = self.locks_path / self.softlock_fname(run_hash)
        if force:
            # Force-release container locks if any
            for container_dir in ('meta', 'seqs'):
                soft_lock_path = self.repo_path / container_dir / 'locks' / self.softlock_fname(run_hash)
                soft_lock_path.unlink(missing_ok=True)
                unix_lock_path = self.repo_path / container_dir / 'locks' / run_hash
                unix_lock_path.unlink(missing_ok=True)

            # Force-release run lock
            lock_path.unlink(missing_ok=True)
        else:
            lock_info = self.get_run_lock_info(run_hash)
            if lock_info.locked and lock_info.version == LockingVersion.LEGACY:
                success = False
            elif lock_info.locked and self.is_stalled_lock(lock_path):
                assert lock_info.version == LockingVersion.NEW
                logger.info(f'Detected stalled lock for Run \'{run_hash}\'. Removing lock.')
                lock_path.unlink()
        return success

    def is_stalled_lock(self, lock_file_path: Path) -> bool:
        with open(lock_file_path, mode='r') as lock_metadata_fh:
            machine_id, pid, *_ = lock_metadata_fh.read().split('-')
            if int(machine_id) == self.machine_id and not psutil.pid_exists(int(pid)):
                return True
        return False



from typing import TYPE_CHECKING, List

from aim.storage.treeutils import encode_tree
from aim.ext.transport.message_utils import pack_args
from aim.ext.transport.remote_resource import RemoteResourceAutoClean

if TYPE_CHECKING:
    from aim.ext.transport.client import Client


class RunProxyAutoClean(RemoteResourceAutoClean):
    PRIORITY = 60


class StructuredRunProxy:
    def __init__(self, client: 'Client',
                 hash_: str,
                 read_only: bool):
        self._resources: RunProxyAutoClean = None
        self._rpc_client = client
        kwargs = {
            'hash_': hash_,
            'read_only': read_only,
        }

        self.init_args = pack_args(encode_tree(kwargs))
        self.resource_type = 'StructuredRun'
        handler = self._rpc_client.get_resource_handler(self, self.resource_type, args=self.init_args)

        self._hash = hash_

        self._resources = RunProxyAutoClean(self)
        self._resources.rpc_client = client
        self._resources.handler = handler
        self._handler = handler

    @property
    def name(self):
        return self._rpc_client.run_instruction(self._hash, self._handler, 'name', [])

    @name.setter
    def name(self, value):
        self._rpc_client.run_instruction(self._hash, self._handler, 'name.setter', (value,), is_write_only=True)

    @property
    def description(self):
        return self._rpc_client.run_instruction(self._hash, self._handler, 'description', [])

    @description.setter
    def description(self, value):
        self._rpc_client.run_instruction(self._hash, self._handler, 'description.setter', (value,), is_write_only=True)

    @property
    def archived(self):
        return self._rpc_client.run_instruction(self._hash, self._handler, 'archived', [])

    @archived.setter
    def archived(self, value):
        self._rpc_client.run_instruction(self._hash, self._handler, 'archived.setter', (value,), is_write_only=True)

    @property
    def creation_time(self):
        return self._rpc_client.run_instruction(self._hash, self._handler, 'creation_time', [])

    @property
    def end_time(self):
        return self._rpc_client.run_instruction(self._hash, self._handler, 'end_time', [])

    @property
    def experiment(self):
        return self._rpc_client.run_instruction(self._hash, self._handler, 'experiment', [])

    @experiment.setter
    def experiment(self, value: str):
        self._rpc_client.run_instruction(self._hash, self._handler, 'experiment.setter', (value,), is_write_only=True)

    @property
    def tags(self) -> List[str]:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'tags', [])

    def add_tag(self, value: str) -> str:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'add_tag', (value,))

    def remove_tag(self, value: str) -> bool:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'remove_tag', (value,))



import pytest

from tests.base import TestBase

from aim.sdk import Repo
from aim.storage.context import Context


class TestRunMetricNumpyTypes(TestBase):
    @pytest.mark.gh_2256
    def test_run_get_metrics(self):
        """covers https://github.com/aimhubio/aim/issues/2256"""
        run = self.create_run(system_tracking_interval=None)
        run_hash = run.hash
        run.track(1.0, name='metric 1', context={})
        run.track(1.0, name='metric 2', context={'metric': 2})
        run.close()

        repo = Repo.default_repo()
        run = repo.get_run(run_hash)
        metric_names = set()
        metric_contexts = set()
        for metric in run.metrics():
            metric_names.add(metric.name)
            metric_contexts.add(metric.context.idx)

        expected_metrics = {'metric 1', 'metric 2'}
        expected_contexts = {Context({}).idx, Context({'metric': 2}).idx}
        self.assertSetEqual(expected_metrics, metric_names)
        self.assertSetEqual(expected_contexts, metric_contexts)



# distutils: language = c++
# cython: language_level = 3

cdef class TreeView:
    cdef:
        object __weakref__

    cpdef view(self, object path, bint resolve = *)
    cpdef object make_array(self, object path = *)
    cpdef object collect(self, object path = *, bint strict = *, bint resolve_objects = *)



import datetime
import pytz

from fastapi import APIRouter as FastAPIRouter
from fastapi import HTTPException
from fastapi.types import DecoratedCallable
from typing import Any, Callable


def object_factory():
    from aim.web.api.projects.project import Project
    project = Project()
    if not project.exists():
        raise HTTPException(status_code=404)

    return project.repo.structured_db


def datetime_now():
    return datetime.datetime.utcnow().replace(tzinfo=pytz.utc)


class APIRouter(FastAPIRouter):
    def api_route(
        self, path: str, *, include_in_schema: bool = True, **kwargs: Any
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        if path.endswith("/"):
            path = path[:-1]

        add_path = super().api_route(
            path, include_in_schema=include_in_schema, **kwargs
        )

        alternate_path = path + "/"
        add_alternate_path = super().api_route(
            alternate_path, include_in_schema=False, **kwargs
        )

        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            add_alternate_path(func)
            return add_path(func)

        return decorator



from tests.base import TestBase

from aim.sdk.run import Run


class TestRunCreationChecks(TestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        run = Run(system_tracking_interval=None, log_system_params=False, capture_terminal_logs=False)
        cls.existing_run_hash = run.hash

    def test_reopen_existing_run_in_write_mode(self):
        run = Run(
            self.existing_run_hash,
            system_tracking_interval=None,
            log_system_params=False,
            capture_terminal_logs=False
        )
        self.assertEqual(run.hash, self.existing_run_hash)

    def test_reopen_existing_run_in_read_mode(self):
        run = Run(
            self.existing_run_hash,
            read_only=True
        )
        self.assertEqual(run.hash, self.existing_run_hash)

    # TODO [deprecation] Uncomment the line below
    # def test_open_non_existing_run_raises_error(self):
    #     from aim.sdk.errors import MissingRunError
    #     with self.assertRaises(MissingRunError):
    #         non_existing_run_hash = '111111111111000000000000'
    #         Run(non_existing_run_hash)



import os

from aim.sdk.repo import Repo
from aim.sdk.configs import get_aim_repo_name
from aim.web.utils import get_root_path


class Project:
    def __init__(self):
        root_path = get_root_path()
        repo_path = f'{root_path}/{get_aim_repo_name()}'

        self.name = 'My awesome project'
        self.path = root_path
        self.repo_path = repo_path
        self.description = ''
        self.repo = Repo.from_path(self.repo_path)
        self.repo.container_pool.clear()
        self.repo.container_view_pool.clear()
        self.repo.persistent_pool.clear()

    def exists(self):
        """
        Checks whether .aim repository is created
        """
        return self.repo and os.path.exists(self.repo_path)



from typing import Optional, List

from aim.cli.upgrade._legacy_repo.repo.trace import Trace


class Metric(object):
    def __init__(self, repo, run, name: str, context: Optional[list] = None):
        self.repo = repo
        self.run = run
        self.name = name
        self.context = context
        self._traces = []  # type: List[Trace]
        self._tmp_all_traces = None  # type: Optional[List[Trace]]
        self._artifact_storage_opened = False

    def __repr__(self):
        return '<{}: {}>\n'.format(self.name, self.traces)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def get_storage(self):
        return self.run.storage

    def open_artifact(self):
        if self._artifact_storage_opened is False:
            self._artifact_storage_opened = True
            self.get_storage().open(self.name, uncommitted_bucket_visible=True)

    def close_artifact(self):
        if self._artifact_storage_opened is True:
            self._artifact_storage_opened = False
            self.get_storage().close(self.name)

    def get_all_traces(self, ignore_empty_context=False) -> list:
        if self._tmp_all_traces:
            return self._tmp_all_traces
        traces = []
        if self.context is None:
            if not ignore_empty_context:
                # Get whole metric data if no context was provided
                traces.append(Trace(self.repo, self, self.name, None))
        else:
            for trace_context in self.context:
                trace = Trace(self.repo, self, self.name, trace_context)
                traces.append(trace)
        self._tmp_all_traces = traces
        return traces



import os.path
from tempfile import TemporaryDirectory

import click

from aim import Run, Image, Text, Audio

IMAGE_EXTENSIONS = ('jpg', 'bmp', 'jpeg', 'png', 'gif', 'svg')
HTML_EXTENSIONS = ('html',)
TEXT_EXTENSIONS = (
    'txt',
    'log',
    'py',
    'js',
    'yaml',
    'yml',
    'json',
    'csv',
    'tsv',
    'md',
    'rst',
    'jsonnet',
)

# Audio is not handled in mlflow but including here just in case
AUDIO_EXTENSIONS = (
    'flac',
    'mp3',
    'wav',
)


def parse_mlflow_logs(repo_inst, tracking_uri, experiment):
    try:
        import mlflow
    except ImportError:
        click.echo(
            'Could not process mlflow logs - failed to import "mlflow" module.', err=True
        )
        return

    client = mlflow.tracking.client.MlflowClient(tracking_uri=tracking_uri)

    if experiment is None:
        # process all experiments
        experiments = client.list_experiments()
    else:
        try:
            ex = client.get_experiment(experiment)
        except mlflow.exceptions.MlflowException:
            ex = client.get_experiment_by_name(experiment)
        if not ex:
            click.echo(f'Could not find experiment with id or name "{experiment}"', err=True)
            return
        experiments = (ex,)

    for ex in experiments:
        runs = client.search_runs(ex.experiment_id)
        for run in runs:
            run_id = run.info.run_id
            aim_run = Run(
                repo=repo_inst,
                system_tracking_interval=None,
                capture_terminal_logs=False,
                experiment=ex.experiment_id,
            )
            aim_run['mlflow_run_id'] = run.info.run_id
            aim_run['mlflow_run_name'] = run.data.tags.get("mlflow.runName")
            aim_run.description = run.data.tags.get("mlflow.note.content")

            # Collect params & tags
            aim_run['params'] = run.data.params
            aim_run['tags'] = {
                k: v for k, v in run.data.tags.items() if not k.startswith('mlflow')
            }

            # Collect metrics
            for key in run.data.metrics.keys():
                for m in client.get_metric_history(run_id, key):
                    aim_run.track(m.value, step=m.step, name=m.key)

            # Collect artifacts
            __html_warning_issued = False
            with TemporaryDirectory(prefix=f'mlflow_{run.info.run_id}_') as temp_path:
                # click.secho(f'Downloading artifacts to {temp_path}', fg='green')
                artifact_loc_stack = [None]
                while artifact_loc_stack:
                    loc = artifact_loc_stack.pop()
                    artifacts = client.list_artifacts(run_id, path=loc)

                    img_batch = []
                    text_batch = []
                    audio_batch = []

                    for file_info in artifacts:
                        if file_info.is_dir:
                            artifact_loc_stack.append(file_info.path)
                            continue

                        downloaded_path = client.download_artifacts(run_id, file_info.path, dst_path=temp_path)
                        if file_info.path.endswith(HTML_EXTENSIONS):
                            # TODO [AP] plotly does not provide interface to load from html
                            # TODO [AP] need to implement html custom object?
                            if not __html_warning_issued:
                                click.secho(
                                    'Handler for html file types is not yet implemented.', fg='yellow'
                                )
                                __html_warning_issued = True
                            continue
                        elif file_info.path.endswith(IMAGE_EXTENSIONS):
                            aim_object = Image
                            kwargs = dict(
                                image=downloaded_path,
                                caption=file_info.path
                            )
                            container = img_batch
                        elif file_info.path.endswith(TEXT_EXTENSIONS):
                            with open(downloaded_path) as fh:
                                content = fh.read()
                            aim_object = Text
                            kwargs = dict(
                                text=content
                            )
                            container = text_batch
                        elif file_info.path.endswith(AUDIO_EXTENSIONS):
                            audio_format = os.path.splitext(file_info.path)[1].lstrip('.')
                            aim_object = Audio
                            kwargs = dict(
                                data=downloaded_path,
                                caption=file_info.path,
                                format=audio_format
                            )
                            container = audio_batch
                        else:
                            click.secho(
                                f'Unresolved or unsupported type for artifact {file_info.path}', fg='yellow'
                            )
                            continue

                        try:
                            item = aim_object(**kwargs)
                        except Exception as exc:
                            click.echo(
                                f'Could not convert artifact {file_info.path} into aim object - {exc}', err=True
                            )
                            continue
                        container.append(item)

                    for content_type, seq in (('image', img_batch),
                                              ('text', text_batch),
                                              ('audio', audio_batch)):
                        aim_run.track(seq, step=0, name=loc or 'root', context={'type': content_type})



import click
from click import core

from aim.cli.configs import *  # noqa F403
from aim.cli.init import commands as init_commands
from aim.cli.version import commands as version_commands
from aim.cli.up import commands as up_commands
from aim.cli.server import commands as server_commands
from aim.cli.reindex import commands as reindex_commands
from aim.cli.runs import commands as runs_commands
from aim.cli.convert import commands as convert_commands
from aim.cli.storage import commands as storage_commands

core._verify_python3_env = lambda: None


@click.group()
@click.option('-v', '--verbose', is_flag=True)
def cli_entry_point(verbose):
    if verbose:
        click.echo('Verbose mode is on')


cli_entry_point.add_command(init_commands.init, INIT_NAME)
cli_entry_point.add_command(version_commands.version, VERSION_NAME)
cli_entry_point.add_command(up_commands.up, UP_NAME)
cli_entry_point.add_command(server_commands.server, SERVER_NAME)
cli_entry_point.add_command(reindex_commands.reindex, REINDEX_NAME)
cli_entry_point.add_command(runs_commands.runs, RUNS_NAME)
cli_entry_point.add_command(convert_commands.convert, CONVERT)
cli_entry_point.add_command(storage_commands.storage, STORAGE)



from typing import Union, Tuple

from aim.sdk.sequence import Sequence
from aim.sdk.objects.figure import Figure


class Figures(Sequence):
    """Class representing series of Plotly figure objects or Plotly lists."""

    @classmethod
    def allowed_dtypes(cls) -> Union[str, Tuple[str, ...]]:
        return (Figure.get_typename(),)  # noqa : need a tuple for consitancy

    @classmethod
    def sequence_name(cls) -> str:
        return "figures"

    def first_step(self):
        """Get sequence tracked first step.

        Required to implement ranged and sliced data fetching.
        """
        return self._meta_tree["first_step"]

    def last_step(self):
        """Get sequence tracked last step.

        Required to implement ranged and sliced data fetching.
        """
        return self._meta_tree["last_step"]



"""initial_schema

Revision ID: 1ecf8222220d
Revises: 
Create Date: 2021-07-26 13:41:56.457649

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1ecf8222220d'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('experiment',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('uuid', sa.Text(), nullable=True),
    sa.Column('name', sa.Text(), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name'),
    sa.UniqueConstraint('name')
    )
    op.create_index(op.f('ix_experiment_uuid'), 'experiment', ['uuid'], unique=True)
    op.create_table('tag',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('uuid', sa.Text(), nullable=True),
    sa.Column('name', sa.Text(), nullable=False),
    sa.Column('color', sa.Text(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name'),
    sa.UniqueConstraint('name')
    )
    op.create_index(op.f('ix_tag_uuid'), 'tag', ['uuid'], unique=True)
    op.create_table('run',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('hash', sa.Text(), nullable=False),
    sa.Column('name', sa.Text(), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.Column('is_archived', sa.Boolean(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('experiment_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['experiment_id'], ['experiment.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_run_hash'), 'run', ['hash'], unique=True)
    op.create_table('run_tag',
    sa.Column('run_id', sa.Integer(), nullable=False),
    sa.Column('tag_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['run_id'], ['run.id'], ),
    sa.ForeignKeyConstraint(['tag_id'], ['tag.id'], ),
    sa.PrimaryKeyConstraint('run_id', 'tag_id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('run_tag')
    op.drop_index(op.f('ix_run_hash'), table_name='run')
    op.drop_table('run')
    op.drop_index(op.f('ix_tag_uuid'), table_name='tag')
    op.drop_table('tag')
    op.drop_index(op.f('ix_experiment_uuid'), table_name='experiment')
    op.drop_table('experiment')
    # ### end Alembic commands ###



from paddle.hapi.callbacks import Callback
from typing import Optional
from aim.sdk.run import Run
from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT


class AimCallback(Callback):
    """
    AimCallback callback function.

    Args:
        repo (:obj:`str`, optional): Aim repository path or Repo object to which Run object is bound.
            If skipped, default Repo is used.
        experiment_name (:obj:`str`, optional): Sets Run's `experiment` property. 'default' if not specified.
            Can be used later to query runs/sequences.
        system_tracking_interval (:obj:`int`, optional): Sets the tracking interval in seconds for system usage
            metrics (CPU, Memory, etc.). Set to `None` to disable system metrics tracking.
        log_system_params (:obj:`bool`, optional): Enable/Disable logging of system params such as installed packages,
            git info, environment variables, etc.
    """

    def __init__(self, repo: Optional[str] = None,
                 experiment_name: Optional[str] = None,
                 system_tracking_interval: Optional[int] = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: bool = True,):
        self.repo = repo
        self.experiment_name = experiment_name
        self.system_tracking_interval = system_tracking_interval
        self.log_system_params = log_system_params
        self._run = None
        self._run_hash = None

    def on_train_begin(self, logs=None):
        self.setup(self.params)

    def on_epoch_begin(self, epoch=None, logs=None):
        self.epoch = epoch

    def on_train_batch_end(self, step, logs=None):
        logs = logs or {}
        self._track(logs, {'subset': 'train'}, step)

    def on_eval_end(self, logs=None):
        self._track(logs, {'subset': 'valid'})

    def _track(self, logs, context, step=None):
        for k, v in logs.items():
            if isinstance(v, list):
                if len(v) == 1:
                    v = v[0]
                else:
                    raise NotImplementedError(f"number of items in {k} are more than 1")
            self._run.track(v, k, step=step, context=context, epoch=self.epoch)

    @property
    def experiment(self):
        if not self._run:
            self.setup()
        return self._run

    def setup(self, args=None):
        if not self._run:
            if self._run_hash:
                self._run = Run(
                    self._run_hash,
                    repo=self.repo,
                    system_tracking_interval=self.system_tracking_interval,
                    log_system_params=self.log_system_params,
                )
            else:
                self._run = Run(
                    repo=self.repo,
                    experiment=self.experiment_name,
                    system_tracking_interval=self.system_tracking_interval,
                    log_system_params=self.log_system_params,
                )
                self._run_hash = self._run.hash

        # Log config parameters
        if args:
            for key in args:
                self._run.set(key, args[key], strict=False)

    def __del__(self):
        if self._run and self._run.active:
            self._run.close()



import logging
import inspect
import traceback
from collections import defaultdict
from typing import Callable, List, Optional, Dict, Any

from aim.sdk.callbacks.helpers import handles_events, get_handler_event_names
logger = logging.getLogger(__name__)


CallbackHandler = Any


class Caller:
    def __init__(self, callbacks: Optional[List['CallbackHandler']] = None):
        self._handlers: Dict[str, List[Callable]] = defaultdict(list)

        if callbacks is None:
            callbacks = []
        for ch in callbacks:
            self.register(ch)

    def register(self, callbacks: 'CallbackHandler'):
        for _, callback in inspect.getmembers(callbacks, handles_events):
            self._bind_events_for(callback)

    def _bind_events_for(self, callback: Callable):
        for e_name in get_handler_event_names(callback):
            self._handlers[e_name].append(callback)

    def _extra_kwargs(self) -> Dict[str, Any]:
        return {'_caller_': self}

    def trigger(self, event_name: str, **kwargs):
        all_kwargs = self._extra_kwargs()
        all_kwargs.update(kwargs)

        handlers = self._handlers.get(event_name, [])
        for handler in handlers:
            try:
                handler(**all_kwargs)
            except Exception:  # noqa
                # TODO catch errors on handler invocation (nice-to-have)
                logger.warning(f'Failed to run callback \'{handler.__name__}\'.')
                logger.warning(traceback.format_exc())



from aim.storage.object import CustomObject


@CustomObject.alias('dvc.metadata')
class DvcData(CustomObject):
    """
    Wrapper over DVC's LIST interface.
    Find DVC tracked files and stores the list into aim storage.
    """
    AIM_NAME = 'dvc.metadata'

    def __init__(self, url='.', path=None, rev=None, recursive=False, dvc_only=False):
        """
        Please refer to DVC reference for kwarg definitions.

        If 'url' is not defined, current dir will be used as DVC directory.
        """

        super().__init__()
        self.storage['dataset'] = {
            'source': 'dvc',
            'tracked_files': self._get_dvc_tracked_files(
                **dict(
                    url=url,
                    path=path,
                    rev=rev,
                    recursive=recursive,
                    dvc_only=dvc_only
                )
            )
        }

    def _get_dvc_tracked_files(self, **ls_kwargs):
        try:
            from dvc.repo import Repo
        except ImportError:
            raise RuntimeError(
                'Could not import module dvc.'
            )

        entries = Repo.ls(**ls_kwargs)
        return [entry['path'] for entry in entries]



import heapq
import logging
import os
import aimrocks

import cachetools.func

from pathlib import Path

from aim.storage.encoding import encode_path
from aim.storage.container import Container, ContainerItemsIterator
from aim.storage.prefixview import PrefixView
from aim.storage.rockscontainer import RocksContainer, optimize_db_for_read

from typing import Dict, List, NamedTuple, Tuple


logger = logging.getLogger(__name__)


class Racer(NamedTuple):
    key: bytes
    priority: int
    value: bytes
    prefix: bytes
    iterator: 'aimrocks.ItemsIterator'


class ItemsIterator(ContainerItemsIterator):
    def __init__(self, dbs: Dict[bytes, "aimrocks.DB"], *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self._iterators = {}
        for key, value in dbs.items():
            self._iterators[key] = value.iteritems(*args, **kwargs)
        self._priority: Dict[bytes, int] = {
            prefix: idx
            for idx, prefix in enumerate(self._iterators)
        }
        self._heap: List[Racer] = []

    def __iter__(self):
        return self

    def __reversed__(self):
        raise NotImplementedError

    def seek_to_first(self):
        for prefix, iterator in self._iterators.items():
            iterator.seek_to_first()
        self._init_heap()

    def seek_to_last(self):
        for prefix, iterator in self._iterators.items():
            iterator.seek_to_last()
        self._init_heap()

    def seek(self, key: bytes):
        for prefix, iterator in self._iterators.items():
            iterator.seek(key)
        self._init_heap()

    def seek_for_prev(self, key):
        for prefix, iterator in self._iterators.items():
            iterator.seek_for_prev(key)
        max_key = self._init_heap()
        self.seek(max_key)

    def get(self) -> Tuple[bytes, bytes]:
        item = self._get(seek_next=False)
        if item is None:
            raise ValueError()
        return item

    def _get(self, seek_next: bool = False) -> Tuple[bytes, bytes]:

        if not self._heap:
            return None

        key, _, value, prefix, iterator = self._heap[0]

        if not seek_next:
            return key, value

        while self._heap:
            alt = self._heap[0]

            if (
                alt.key != key and (alt.prefix == prefix or not alt.key.startswith(prefix))
            ):
                break

            heapq.heappop(self._heap)
            try:
                new_key, new_value = next(alt.iterator)
                racer = Racer(new_key, alt.priority, new_value, alt.prefix, alt.iterator)
                heapq.heappush(self._heap, racer)
                # self._state[prefix] = (new_key, new_value)
            except StopIteration:
                pass

        return key, value

    def _init_heap(self):
        self._heap = []
        max_key = b''
        for prefix, iterator in self._iterators.items():
            try:
                key, value = iterator.get()
                max_key = max(key, max_key)
            except ValueError:
                continue
            racer = Racer(key, self._priority[prefix], value, prefix, iterator)
            heapq.heappush(self._heap, racer)

        return max_key

    def next(self) -> Tuple[bytes, bytes]:
        return self._get(seek_next=True)


# We temporary implement KeysIterator and ValuesIterator
# using ItemsIterator because the performance is not better
# when iterating over only one of them

class KeysIterator(ItemsIterator):

    def get(self) -> bytes:
        key, value = super().get()
        return key

    def __next__(self) -> bytes:
        key, value = super().__next__()
        return key


class ValuesIterator(ItemsIterator):
    def get(self) -> bytes:
        key, value = super().get()
        return value

    def __next__(self) -> bytes:
        key, value = super().__next__()
        return value


class DB(object):
    def __init__(self, db_path: str, db_name: str, opts, read_only: bool = False):
        assert read_only
        self.db_path = db_path
        self.db_name = db_name
        self.opts = opts
        self._dbs: Dict[bytes, aimrocks.DB] = dict()

    def _get_db(
        self,
        prefix: bytes,
        path: str,
        cache: Dict[bytes, aimrocks.DB],
        store: Dict[bytes, aimrocks.DB] = None,
    ):
        db = cache.get(prefix)
        if db is None:
            optimize_db_for_read(Path(path), self.opts)
            db = aimrocks.DB(path, opts=aimrocks.Options(**self.opts), read_only=True)
        if store is not None:
            store[prefix] = db
        return db

    @cachetools.func.ttl_cache(maxsize=None, ttl=0.1)
    def _list_dir(self, path: str):
        try:
            return os.listdir(path)
        except FileNotFoundError:
            return []

    @property
    @cachetools.func.ttl_cache(maxsize=None, ttl=0.1)
    def dbs(self):
        index_prefix = encode_path((self.db_name, "chunks"))
        index_path = os.path.join(self.db_path, self.db_name, "index")
        try:
            index_db = self._get_db(index_prefix, index_path, self._dbs)
        except Exception:
            index_db = None
            logger.info('No index was detected')

        # If index exists -- only load those in progress
        selector = 'progress' if index_db is not None else 'chunks'

        new_dbs: Dict[bytes, aimrocks.DB] = {}
        db_dir = os.path.join(self.db_path, self.db_name, selector)
        for prefix in self._list_dir(db_dir):
            path = os.path.join(self.db_path, self.db_name, "chunks", prefix)
            prefix = encode_path((self.db_name, "chunks", prefix))
            self._get_db(prefix, path, self._dbs, new_dbs)

        if index_db is not None:
            new_dbs[b""] = index_db
        self._dbs = new_dbs
        return new_dbs

    def close(self):
        ...

    def get(self, key: bytes, *args, **kwargs) -> bytes:
        for prefix, db in self.dbs.items():
            # Shadowing
            if key.startswith(prefix):
                return db.get(key)
        return self.dbs[b""].get(key)

    def iteritems(
        self, *args, **kwargs
    ) -> "ItemsIterator":
        return ItemsIterator(self.dbs, *args, **kwargs)

    def iterkeys(
        self, *args, **kwargs
    ) -> "KeysIterator":
        return KeysIterator(self.dbs, *args, **kwargs)

    def itervalues(
        self, *args, **kwargs
    ) -> "ValuesIterator":
        return ValuesIterator(self.dbs, *args, **kwargs)


class RocksUnionContainer(RocksContainer):

    def __init__(self, *args, **kwargs):
        return super().__init__(*args, **kwargs)

    @property
    def writable_db(self) -> aimrocks.DB:
        raise NotImplementedError

    @property
    def db(self) -> aimrocks.DB:
        assert self.read_only

        if self._db is not None:
            return self._db
        try:
            logger.debug(f'opening {self.path} as aimrocks db')
            path = Path(self.path)
            path.parent.mkdir(parents=True, exist_ok=True)

            self._db = DB(str(path.parent), path.name, self._db_opts, read_only=self.read_only)
        except Exception as e:
            # print(e, self.path)
            raise e

        return self._db

    def view(
        self,
        prefix: bytes = b''
    ) -> 'Container':
        container = self
        if prefix and prefix in self.db.dbs:
            container = RocksUnionSubContainer(container=self, domain=prefix)
        return PrefixView(prefix=prefix,
                          container=container)


class RocksUnionSubContainer(RocksContainer):
    def __init__(self, container: 'RocksUnionContainer', domain: bytes):
        self._parent = container
        self.domain = domain

    @property
    def writable_db(self) -> aimrocks.DB:
        raise NotImplementedError

    @property
    def db(self) -> aimrocks.DB:
        db: DB = self._parent.db
        return db.dbs.get(self.domain, db)

    def view(
        self,
        prefix: bytes = b''
    ) -> 'Container':
        return PrefixView(prefix=prefix, container=self)



"""empty message

Revision ID: 9ba30ab3b2b4
Revises: fbfe5c4702fb
Create Date: 2021-08-27 18:17:27.819214

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9ba30ab3b2b4'
down_revision = 'fbfe5c4702fb'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('tag', sa.Column('description', sa.Text(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('tag', 'description')
    # ### end Alembic commands ###
